"""Utility for determining the type of an image from its bytes."""


def get_image_type(image_data: bytes) -> str:
    if image_data.startswith(b"\xff\xd8\xff"):
        return "jpeg"
    elif image_data.startswith(b"\x89PNG\r\n\x1a\n"):
        return "png"
    elif image_data.startswith(b"GIF87a") or image_data.startswith(b"GIF89a"):
        return "gif"
    elif image_data.startswith(b"RIFF") and image_data[8:12] == b"WEBP":
        return "webp"
    elif image_data[4:12] in (
        b"ftypmif1",
        b"ftypmsf1",
        b"ftypheic",
        b"ftypheix",
        b"ftyphevc",
        b"ftyphevx",
    ):
        # HEIF and HEIC files start with 'ftyp' followed by specific codes
        if image_data[4:8] == b"ftyp":
            subtype = image_data[8:12]
            if subtype in (b"heic", b"heix"):
                return "heic"
            elif subtype in (b"mif1", b"msf1", b"hevc", b"hevx"):
                return "heif"
    raise ValueError("Unsupported image type")



"""The `groq_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import GroqCallParams
from .call_response import GroqCallResponse
from .call_response_chunk import GroqCallResponseChunk
from .dynamic_config import GroqDynamicConfig
from .stream import GroqStream
from .tool import GroqTool

groq_call = call_factory(
    TCallResponse=GroqCallResponse,
    TCallResponseChunk=GroqCallResponseChunk,
    TDynamicConfig=GroqDynamicConfig,
    TToolType=GroqTool,
    TStream=GroqStream,
    TCallParams=GroqCallParams,
    default_call_params=GroqCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the Groq API with a typed function.

This decorator is used to wrap a typed function that calls the Groq API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@groq_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The Groq model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the Groq API call.
    **call_params: The `GroqCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an Groq API call.
'''



"""This module contains the setup_call function for the Anthropic API."""

import inspect
from typing import Any, Awaitable, Callable, cast

from anthropic import (
    Anthropic,
    AnthropicBedrock,
    AnthropicVertex,
    AsyncAnthropic,
    AsyncAnthropicBedrock,
    AsyncAnthropicVertex,
)
from anthropic.types import Message, MessageParam

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import AnthropicCallParams
from ..dynamic_config import AnthropicDynamicConfig
from ..tool import AnthropicTool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: Anthropic
    | AsyncAnthropic
    | AnthropicBedrock
    | AsyncAnthropicBedrock
    | AnthropicVertex
    | AsyncAnthropicVertex
    | None,
    fn: Callable[..., AnthropicDynamicConfig | Awaitable[AnthropicDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AnthropicDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AnthropicCallParams,
    extract: bool,
) -> tuple[
    Callable[..., Message] | Callable[..., Awaitable[Message]],
    str,
    list[MessageParam],
    list[type[AnthropicTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, AnthropicTool, call_params
    )
    messages = cast(list[BaseMessageParam | MessageParam], messages)
    messages = convert_message_params(messages)

    if messages[0]["role"] == "system":
        call_kwargs["system"] = messages.pop(0)["content"]

    if json_mode:
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        if isinstance(messages[-1]["content"], str):
            messages[-1]["content"] += json_mode_content
        else:
            messages[-1]["content"] = list(messages[-1]["content"]) + [
                {"type": "text", "text": json_mode_content}
            ]
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = {"type": "tool", "name": tool_types[0]._name()}
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = AsyncAnthropic() if inspect.iscoroutinefunction(fn) else Anthropic()
    create = client.messages.create
    return create, prompt_template, messages, tool_types, call_kwargs



"""This module contains the `convert_base_type_to_base_tool` function."""

from typing import Annotated, TypeVar, get_args, get_origin

from pydantic import BaseModel, create_model

from ._base_type import BaseType
from ._default_tool_docstring import DEFAULT_TOOL_DOCSTRING

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def convert_base_type_to_base_tool(
    schema: type[BaseType], base: type[BaseToolT]
) -> type[BaseToolT]:
    """Converts a `BaseType` to a `BaseToolT` type."""
    if get_origin(schema) == Annotated:
        schema.__name__ = get_args(schema)[0].__name__
    return create_model(
        schema.__name__,
        __base__=base,
        __doc__=DEFAULT_TOOL_DOCSTRING,
        value=(schema, ...),
    )



"""This module contains the `GeminiCallResponse` class."""

from typing import Optional

from google.generativeai.protos import FunctionResponse  # type: ignore
from google.generativeai.types import (  # type: ignore
    AsyncGenerateContentResponse,
    ContentDict,
    ContentsType,  # type: ignore
    GenerateContentResponse,
)
from pydantic import computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import GeminiCallParams
from .dynamic_config import GeminiDynamicConfig
from .tool import GeminiTool


class GeminiCallResponse(
    BaseCallResponse[
        GenerateContentResponse | AsyncGenerateContentResponse,
        GeminiTool,
        GeminiDynamicConfig,
        ContentsType,
        GeminiCallParams,
        ContentDict,
    ]
):
    '''Convenience wrapper around Gemini's `GenerateContentResponse`.

    When using Mirascope's convenience wrappers to interact with Gemini models via
    `GeminiCall`, responses using `GeminiCall.call()` will return a
    `GeminiCallResponse`, whereby the implemented properties allow for simpler syntax
    and a convenient developer experience.

    Example:

    ```python
    from mirascope.core.gemini import gemini_call

    @gemini_call(model="gemini-1.5-pro")
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    response = recommend_book("fantasy")  # response is an `GeminiCallResponse` instance
    print(response.content)
    #> Sure! I would recommend...
    ```
    '''

    _provider = "gemini"

    @property
    def content(self) -> str:
        """Returns the contained string content for the 0th choice."""
        return self.response.candidates[0].content.parts[0].text

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        finish_reasons = [
            "FINISH_REASON_UNSPECIFIED",
            "STOP",
            "MAX_TOKENS",
            "SAFETY",
            "RECITATION",
            "OTHER",
        ]

        return [
            finish_reasons[candidate.finish_reason]
            for candidate in self.response.candidates
        ]

    @property
    def model(self) -> str:
        """Returns the model name.

        google.generativeai does not return model, so we return the model provided by
        the user.
        """
        return self._model

    @property
    def id(self) -> Optional[str]:
        """Returns the id of the response.

        google.generativeai does not return an id
        """
        return None

    @property
    def usage(self) -> None:
        """Returns the usage of the chat completion.

        google.generativeai does not have Usage, so we return None
        """
        return None

    @property
    def input_tokens(self) -> None:
        """Returns the number of input tokens."""
        return None

    @property
    def output_tokens(self) -> None:
        """Returns the number of output tokens."""
        return None

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> ContentDict:
        """Returns the models's response as a message parameter."""
        return {"role": "model", "parts": self.response.parts}  # type: ignore

    @computed_field
    @property
    def tools(self) -> list[GeminiTool] | None:
        """Returns the list of tools for the 0th candidate's 0th content part."""
        if self.tool_types is None:
            return None

        extracted_tools = []
        for part in self.response.candidates[0].content.parts:
            tool_call = part.function_call
            for tool_type in self.tool_types:
                if tool_call.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> GeminiTool | None:
        """Returns the 0th tool for the 0th candidate's 0th content part.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        tools = self.tools
        if tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[GeminiTool, object]]
    ) -> list[FunctionResponse]:
        """Returns the tool message parameters for tool call results."""
        return [
            FunctionResponse(name=tool._name(), response={"result": output})
            for tool, output in tools_and_outputs
        ]



"""Get the JSON output from an anthropic call response or chunk."""

import json

from ..call_response import AnthropicCallResponse
from ..call_response_chunk import AnthropicCallResponseChunk


def get_json_output(
    response: AnthropicCallResponse | AnthropicCallResponseChunk, json_mode: bool
) -> str:
    """Get the JSON output from a completion response."""
    if isinstance(response, AnthropicCallResponse):
        if json_mode and (content := response.content):
            json_start = content.index("{")
            json_end = content.rfind("}")
            return content[json_start : json_end + 1]
        elif (
            (block := response.response.content[0])
            and block.type == "tool_use"
            and block.input is not None
        ):
            return json.dumps(block.input)
        raise ValueError("No tool call or JSON object found in response.")
    else:
        if json_mode:
            return response.content
        elif (
            response.chunk.type == "content_block_delta"
            and (delta := response.chunk.delta)
            and delta.type == "input_json_delta"
        ):
            return delta.partial_json
        return ""



from typing import Any, Literal, NotRequired, Sequence

from typing_extensions import TypedDict

from ..base.embedding_params import BaseEmbeddingParams


class RequestOptions(TypedDict):
    """Redefining their class to use `typing_extensions.TypedDict` for Pydantic."""

    timeout_in_seconds: NotRequired[int]
    max_retries: NotRequired[int]
    additional_headers: NotRequired[dict[str, Any]]
    additional_query_parameters: NotRequired[dict[str, Any]]
    additional_body_parameters: NotRequired[dict[str, Any]]


class CohereEmbeddingParams(BaseEmbeddingParams):
    model: str = "embed-english-v3.0"
    input_type: Literal[
        "search_document", "search_query", "classification", "clustering"
    ] = "search_query"
    embedding_types: (
        Sequence[Literal["float", "int8", "uint8", "binary", "ubinary"]] | None
    ) = None
    truncate: Literal["none", "end", "start"] | None = "end"
    request_options: RequestOptions | None = None
    batching: bool | None = True



"""The `mistral_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import MistralCallParams
from .call_response import MistralCallResponse
from .call_response_chunk import MistralCallResponseChunk
from .dynamic_config import MistralDynamicConfig
from .stream import MistralStream
from .tool import MistralTool

mistral_call = call_factory(
    TCallResponse=MistralCallResponse,
    TCallResponseChunk=MistralCallResponseChunk,
    TDynamicConfig=MistralDynamicConfig,
    TToolType=MistralTool,
    TStream=MistralStream,
    TCallParams=MistralCallParams,
    default_call_params=MistralCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the Mistral API with a typed function.

This decorator is used to wrap a typed function that calls the Mistral API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@mistral_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The Mistral model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the Mistral API call.
    **call_params: The `MistralCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an Mistral API call.
'''



"""This module contains the setup_call function for Mistral tools."""

import inspect
from typing import Any, Awaitable, Callable, cast

from mistralai.async_client import MistralAsyncClient
from mistralai.client import MistralClient
from mistralai.models.chat_completion import (
    ChatCompletionResponse,
    ChatMessage,
    ResponseFormat,
    ResponseFormats,
    ToolChoice,
)

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import MistralCallParams
from ..dynamic_config import MistralDynamicConfig
from ..tool import MistralTool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: MistralClient | MistralAsyncClient | None,
    fn: Callable[..., MistralDynamicConfig | Awaitable[MistralDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: MistralDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: MistralCallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletionResponse]
    | Callable[..., Awaitable[ChatCompletionResponse]],
    str,
    list[ChatMessage],
    list[type[MistralTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, MistralTool, call_params
    )
    messages = cast(list[BaseMessageParam | ChatMessage], messages)
    messages = convert_message_params(messages)
    if json_mode:
        call_kwargs["response_format"] = ResponseFormat(
            type=ResponseFormats("json_object")
        )
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        if messages[-1].role == "user":
            messages[-1].content += json_mode_content
        else:
            messages.append(ChatMessage(role="user", content=json_mode_content.strip()))
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = ToolChoice.any
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = (
            MistralAsyncClient() if inspect.iscoroutinefunction(fn) else MistralClient()
        )

    def create_or_stream(stream: bool, **kwargs: Any):
        if stream:
            return client.chat_stream(**kwargs)
        return client.chat(**kwargs)

    return create_or_stream, prompt_template, messages, tool_types, call_kwargs  # type: ignore



"""The `cohere_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import CohereCallParams
from .call_response import CohereCallResponse
from .call_response_chunk import CohereCallResponseChunk
from .dynamic_config import CohereDynamicConfig
from .stream import CohereStream
from .tool import CohereTool

cohere_call = call_factory(
    TCallResponse=CohereCallResponse,
    TCallResponseChunk=CohereCallResponseChunk,
    TDynamicConfig=CohereDynamicConfig,
    TToolType=CohereTool,
    TStream=CohereStream,
    TCallParams=CohereCallParams,
    default_call_params=CohereCallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the Cohere API with a typed function.

This decorator is used to wrap a typed function that calls the Cohere API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@cohere_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The Cohere model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the Cohere API call.
    **call_params: The `CohereCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an Cohere API call.
'''



"""Mirascope x OpenTelemetry Integration."""

from ..middleware_factory import middleware_decorator
from ._utils import (
    custom_context_manager,
    handle_call_response,
    handle_call_response_async,
    handle_response_model,
    handle_response_model_async,
    handle_stream,
    handle_stream_async,
    handle_structured_stream,
    handle_structured_stream_async,
)


def with_otel(fn):
    """Wraps a Mirascope function with OpenTelemetry."""

    return middleware_decorator(
        fn,
        custom_context_manager=custom_context_manager,
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_response_model=handle_response_model,
        handle_response_model_async=handle_response_model_async,
        handle_structured_stream=handle_structured_stream,
        handle_structured_stream_async=handle_structured_stream_async,
    )



"""The `CohereTool` class for easy tool usage with Cohere LLM calls."""

from __future__ import annotations

from cohere.types import Tool, ToolCall, ToolParameterDefinitionsValue
from pydantic import SkipValidation
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class CohereTool(BaseTool):
    """A class for defining tools for Cohere LLM calls."""

    tool_call: SkipValidation[SkipJsonSchema[ToolCall]]

    @classmethod
    def tool_schema(cls) -> Tool:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined."""
        model_schema = cls.model_tool_schema()
        parameter_definitions = None
        if "properties" in model_schema:
            if "$defs" in model_schema["properties"]:
                raise ValueError(  # pragma: no cover
                    "Unfortunately Cohere's chat API cannot handle nested structures "
                    "with $defs."
                )
            parameter_definitions = {
                prop: ToolParameterDefinitionsValue(
                    description=prop_schema["description"]
                    if "description" in prop_schema
                    else None,
                    type=prop_schema["type"],
                    required="required" in model_schema
                    and prop in model_schema["required"],
                )
                for prop, prop_schema in model_schema["properties"].items()
            }
        return Tool(
            name=cls._name(),
            description=cls._description(),
            parameter_definitions=parameter_definitions,
        )

    @classmethod
    def from_tool_call(cls, tool_call: ToolCall) -> CohereTool:
        """Constructs an `CohereTool` instance from a `tool_call`."""
        model_json = {**tool_call.parameters}
        model_json["tool_call"] = tool_call.dict()
        return cls.model_validate(model_json)



"""The `MistralStream` class for convenience around streaming LLM calls."""

from mistralai.models.chat_completion import (
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    ChatMessage,
    FinishReason,
)
from mistralai.models.common import UsageInfo

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import MistralCallParams
from .call_response import MistralCallResponse
from .call_response_chunk import MistralCallResponseChunk
from .dynamic_config import MistralDynamicConfig
from .tool import MistralTool


class MistralStream(
    BaseStream[
        MistralCallResponse,
        MistralCallResponseChunk,
        ChatMessage,
        ChatMessage,
        ChatMessage,
        ChatMessage,
        MistralTool,
        MistralDynamicConfig,
        MistralCallParams,
        FinishReason,
    ]
):
    _provider = "mistral"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self, tool_calls: list | None = None, content: str | None = None
    ) -> ChatMessage:
        message_param = ChatMessage(
            role="assistant", content=content if content else "", tool_calls=tool_calls
        )
        return message_param

    def construct_call_response(self) -> MistralCallResponse:
        """Constructs the call response from a consumed MistralStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        usage = UsageInfo(
            prompt_tokens=int(self.input_tokens or 0),
            completion_tokens=int(self.output_tokens or 0),
            total_tokens=int(self.input_tokens or 0) + int(self.output_tokens or 0),
        )
        completion = ChatCompletionResponse(
            id=self.id if self.id else "",
            choices=[
                ChatCompletionResponseChoice(
                    finish_reason=self.finish_reasons[0]
                    if self.finish_reasons
                    else None,
                    index=0,
                    message=self.message_param,
                )
            ],
            created=0,
            model=self.model,
            object="",
            usage=usage,
        )
        return MistralCallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



"""Utility for determining the type of an audio file from its bytes."""


def get_audio_type(audio_data: bytes) -> str:
    if audio_data.startswith(b"RIFF") and audio_data[8:12] == b"WAVE":
        return "wav"
    elif audio_data.startswith(b"ID3") or audio_data.startswith(b"\xff\xfb"):
        return "mp3"
    elif audio_data.startswith(b"FORM") and audio_data[8:12] == b"AIFF":
        return "aiff"
    elif audio_data.startswith(b"\xff\xf1") or audio_data.startswith(b"\xff\xf9"):
        return "aac"
    elif audio_data.startswith(b"OggS"):
        return "ogg"
    elif audio_data.startswith(b"fLaC"):
        return "flac"

    raise ValueError("Unsupported audio type")



"""This module contains the `BaseCallResponseChunk` class."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, ConfigDict

_ChunkT = TypeVar("_ChunkT", bound=Any)
_FinishReasonT = TypeVar("_FinishReasonT", bound=Any)


class BaseCallResponseChunk(BaseModel, Generic[_ChunkT, _FinishReasonT], ABC):
    """A base abstract interface for LLM streaming response chunks.

    Attributes:
        chunk: The original response chunk from whichever model response this wraps.
        tool_types: The tool types sent in the LLM call if any.
        user_message_param: The most recent message if it was a user message. Otherwise
            `None`.
        cost: The cost of the completion in dollars.
    """

    chunk: _ChunkT

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __str__(self) -> str:
        """Returns the string content of the chunk."""
        return self.content

    @property
    @abstractmethod
    def content(self) -> str:
        """Should return the string content of the response chunk.

        If there are multiple choices in a chunk, this method should select the 0th
        choice and return it's string content.

        If there is no string content (e.g. when using tools), this method must return
        the empty string.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def finish_reasons(self) -> list[_FinishReasonT] | None:
        """Should return the finish reasons of the response.

        If there is no finish reason, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def model(self) -> str | None:
        """Should return the name of the response model."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def id(self) -> str | None:
        """Should return the id of the response."""
        ...  # pragma: no cover

    @property
    @abstractmethod
    def usage(self) -> Any:
        """Should return the usage of the response.

        If there is no usage, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def input_tokens(self) -> int | float | None:
        """Should return the number of input tokens.

        If there is no input_tokens, this method must return None.
        """
        ...  # pragma: no cover

    @property
    @abstractmethod
    def output_tokens(self) -> int | float | None:
        """Should return the number of output tokens.

        If there is no output_tokens, this method must return None.
        """
        ...  # pragma: no cover



"""This module contains the setup_call function for OpenAI tools."""

import inspect
from typing import Any, Awaitable, Callable, cast

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import OpenAICallParams
from ..dynamic_config import OpenAIDynamicConfig
from ..tool import OpenAITool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: OpenAI | AsyncOpenAI | AzureOpenAI | AsyncAzureOpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
    str,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, OpenAITool, call_params
    )
    messages = cast(list[BaseMessageParam | ChatCompletionMessageParam], messages)
    messages = convert_message_params(messages)
    if json_mode:
        call_kwargs["response_format"] = {"type": "json_object"}
        json_mode_content = _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        messages.append(
            ChatCompletionUserMessageParam(role="user", content=json_mode_content)
        )
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs["tool_choice"] = "required"
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        client = AsyncOpenAI() if inspect.iscoroutinefunction(fn) else OpenAI()
    create = client.chat.completions.create

    return create, prompt_template, messages, tool_types, call_kwargs



"""This module provides a function to get the variables in a template string."""

from string import Formatter
from typing import Literal, overload


@overload
def get_template_variables(
    template: str, include_format_spec: Literal[True]
) -> list[tuple[str, str | None]]: ...  # pragma: no cover


@overload
def get_template_variables(
    template: str, include_format_spec: Literal[False]
) -> list[str]: ...  # pragma: no cover


def get_template_variables(
    template: str, include_format_spec: bool
) -> list[str] | list[tuple[str, str | None]]:
    """Returns the variables in the given template string.

    Args:
        template: The template string to parse.

    Returns:
        The variables in the template string.
    """
    if include_format_spec:
        return [
            (var, format_spec)
            for _, var, format_spec, _ in Formatter().parse(template)
            if var
        ]
    else:
        return [var for _, var, _, _ in Formatter().parse(template) if var]



"""Internal Utilities."""

from ._base_type import BaseType, is_base_type
from ._convert_base_model_to_base_tool import convert_base_model_to_base_tool
from ._convert_base_type_to_base_tool import convert_base_type_to_base_tool
from ._convert_function_to_base_tool import convert_function_to_base_tool
from ._default_tool_docstring import DEFAULT_TOOL_DOCSTRING
from ._extract_tool_return import extract_tool_return
from ._format_template import format_template
from ._get_audio_type import get_audio_type
from ._get_fn_args import get_fn_args
from ._get_image_type import get_image_type
from ._get_metadata import get_metadata
from ._get_possible_user_message_param import get_possible_user_message_param
from ._get_prompt_template import get_prompt_template
from ._get_template_values import get_template_values
from ._get_template_variables import get_template_variables
from ._json_mode_content import json_mode_content
from ._parse_content_template import parse_content_template
from ._parse_prompt_messages import parse_prompt_messages
from ._protocols import (
    AsyncCreateFn,
    CalculateCost,
    CreateFn,
    GetJsonOutput,
    HandleStream,
    HandleStreamAsync,
    LLMFunctionDecorator,
    SetupCall,
)
from ._setup_call import setup_call
from ._setup_extract_tool import setup_extract_tool

__all__ = [
    "AsyncCreateFn",
    "BaseType",
    "CalculateCost",
    "convert_base_model_to_base_tool",
    "convert_base_type_to_base_tool",
    "convert_function_to_base_tool",
    "CreateFn",
    "DEFAULT_TOOL_DOCSTRING",
    "extract_tool_return",
    "format_template",
    "GetJsonOutput",
    "get_audio_type",
    "get_fn_args",
    "get_image_type",
    "get_metadata",
    "get_possible_user_message_param",
    "get_prompt_template",
    "get_template_values",
    "get_template_variables",
    "HandleStream",
    "HandleStreamAsync",
    "is_base_type",
    "json_mode_content",
    "LLMFunctionDecorator",
    "parse_content_template",
    "parse_prompt_messages",
    "SetupCall",
    "setup_call",
    "setup_extract_tool",
]



"""Utility for pulling metadata from a call and merging with any dynamic metadata."""

from typing import Callable

from pydantic import BaseModel

from ..dynamic_config import BaseDynamicConfig
from ..metadata import Metadata


def get_metadata(
    fn: Callable | BaseModel, dynamic_config: BaseDynamicConfig
) -> Metadata:
    """Get the metadata from the function and merge with any dynamic metadata."""
    if dynamic_config and "metadata" in dynamic_config:
        return dynamic_config["metadata"]
    return getattr(fn, "_metadata", Metadata())



from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar, Union

from pydantic import BaseModel, ConfigDict

ResponseT = TypeVar("ResponseT", bound=Any)


class BaseEmbeddingResponse(BaseModel, Generic[ResponseT], ABC):
    """A base abstract interface for LLM embedding responses.

    Attributes:
        response: The original response from whichever model response this wraps.
    """

    response: ResponseT
    start_time: float  # The start time of the embedding in ms
    end_time: float  # The end time of the embedding in ms

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    @property
    @abstractmethod
    def embeddings(self) -> Optional[Union[list[list[float]], list[list[int]]]]:
        """Should return the embedding of the response.

        If there are multiple choices in a response, this method should select the 0th
        choice and return it's embedding.
        """
        ...  # pragma: no cover



"""The `anthropic_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import AnthropicCallParams
from .call_response import AnthropicCallResponse
from .call_response_chunk import AnthropicCallResponseChunk
from .dynamic_config import AnthropicDynamicConfig
from .stream import AnthropicStream
from .tool import AnthropicTool

anthropic_call = call_factory(
    TCallResponse=AnthropicCallResponse,
    TCallResponseChunk=AnthropicCallResponseChunk,
    TDynamicConfig=AnthropicDynamicConfig,
    TStream=AnthropicStream,
    TToolType=AnthropicTool,
    TCallParams=AnthropicCallParams,
    default_call_params=AnthropicCallParams(max_tokens=1000),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the Anthropic API with a typed function.

This decorator is used to wrap a typed function that calls the Anthropic API. It
parses the docstring of the wrapped function as the messages array and templates the
input arguments for the function into each message's template.

Example:

```python
@anthropic_call(model="claude-3-5-sonnet-20240620")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The Anthropic model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the Anthropic API call.
    **call_params: The `AnthropicCallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an Anthropic API call.
'''



"""The module for defining the toolkit class for LLM call tools."""

from __future__ import annotations

import inspect
from abc import ABC
from typing import Callable, ClassVar, NamedTuple, TypeVar

from pydantic import BaseModel, ConfigDict
from typing_extensions import Concatenate, ParamSpec

from . import BaseTool
from ._utils import convert_function_to_base_tool, get_template_variables

_TOOLKIT_TOOL_METHOD_MARKER: str = "__toolkit_tool_method__"

_namespaces: set[str] = set()

P = ParamSpec("P")


class ToolKitToolMethod(NamedTuple):
    method: Callable[..., str]
    template_vars: list[str]
    template: str


class BaseToolKit(BaseModel, ABC):
    """A class for defining tools for LLM call tools.

    The class should have methods decorated with `@toolkit_tool` to create tools.

    Example:
    ```python
    from mirascope.core.base import BaseToolKit, toolkit_tool
    from mirascope.core import openai

    class BookRecommendationToolKit(BaseToolKit):
        '''A toolkit for recommending books.'''

        __namespace__: ClassVar[str | None] = 'book_tools'
        reading_level: Literal["beginner", "advanced"]

        @toolkit_tool
        def format_book(self, title: str, author: str) -> str:
            '''Returns the title and author of a book nicely formatted.

            Reading level: {self.reading_level}
            '''
            return f"{title} by {author}"

    @openai.call(model="gpt-4o")
    def recommend_book(genre: str, reading_level: Literal["beginner", "advanced"]):
        '''Recommend a {genre} book.'''
        toolkit = BookRecommendationToolKit(reading_level=reading_level)
        return {"tools": toolkit.create_tools()}

    response = recommend_book("fantasy", "beginner")
    if tool := response.tool:
        output = tool.call()
        print(output)
        #> The Name of the Wind by Patrick Rothfuss
    else:
        print(response.content)
        #> Sure! I would recommend...
    ```
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    _toolkit_tool_methods: ClassVar[list[ToolKitToolMethod]]
    __namespace__: ClassVar[str | None] = None

    def create_tools(self) -> list[type[BaseTool]]:
        """The method to create the tools."""
        tools = []
        for method, template_vars, template in self._toolkit_tool_methods:
            for var in template_vars:
                if var.startswith("self."):
                    continue
                # Replace non-self template variables with escaped double brackets so
                # that templating `self` results in a future templateable string.
                template = template.replace(f"{{{var}}}", f"{{{{{var}}}}}")
            tools.append(
                convert_function_to_base_tool(
                    method, BaseTool, template.format(self=self), self.__namespace__
                )
            )
        return tools

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs):
        # validate the namespace
        if cls.__namespace__:
            if cls.__namespace__ in _namespaces:
                raise ValueError(f"The namespace {cls.__namespace__} is already used")
            _namespaces.add(cls.__namespace__)

        cls._toolkit_tool_methods = []
        for attr in cls.__dict__.values():
            if not getattr(attr, _TOOLKIT_TOOL_METHOD_MARKER, False):
                continue
            # Validate the toolkit_tool_method
            if (template := attr.__doc__) is None:
                raise ValueError("The toolkit_tool method must have a docstring")

            dedented_template = inspect.cleandoc(template)
            template_vars = get_template_variables(dedented_template, False)

            for var in template_vars:
                if not var.startswith("self."):
                    # Skip template variables that do not start with self as they may
                    # be later templated e.g. if using a call as a tool.
                    continue

                self_var = var[5:]

                # Expecting pydantic model fields or class attribute and property
                if self_var in cls.model_fields or hasattr(cls, self_var):
                    continue
                raise ValueError(
                    f"The toolkit_tool method template variable {var} is not found in "
                    "the class"
                )

            cls._toolkit_tool_methods.append(
                ToolKitToolMethod(attr, template_vars, dedented_template)
            )
        if not cls._toolkit_tool_methods:
            raise ValueError("No toolkit_tool method found")


_BaseToolKitT = TypeVar("_BaseToolKitT", bound=BaseToolKit)


def toolkit_tool(
    method: Callable[Concatenate[_BaseToolKitT, P], str],
) -> Callable[Concatenate[_BaseToolKitT, P], str]:
    # Mark the method as a toolkit tool
    setattr(method, _TOOLKIT_TOOL_METHOD_MARKER, True)

    return method



"""Function for binding `args` and `kwargs` as a dictionary to the fn's signature."""

import inspect
from typing import Any, Callable


def get_fn_args(
    fn: Callable, args: tuple[object, ...], kwargs: dict[str, Any]
) -> dict[str, Any]:
    """Returns the `args` and `kwargs` as a dictionary bound by `fn`'s signature."""
    bound_args = inspect.signature(fn).bind_partial(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments



"""A function generating content to request JSON mode from models without it."""

import json

from ..tool import BaseTool


def json_mode_content(tool_type: type[BaseTool] | None):
    """Returns the content to request JSON mode from models without it."""
    if not tool_type:
        return "\n\nExtract ONLY a valid JSON dict using the schema."
    return f"""

Extract ONLY a valid JSON dict (NOT THE SCHEMA) from the content that adheres to this schema:
{json.dumps(tool_type.model_json_schema(), indent=2)}"""



"""A module for interacting with Cohere chat models."""

from .embedders import CohereEmbedder
from .embedding_params import CohereEmbeddingParams
from .embedding_response import CohereEmbeddingResponse

__all__ = [
    "CohereEmbedder",
    "CohereEmbeddingParams",
    "CohereEmbeddingResponse",
]



"""The `litellm_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ..openai import (
    OpenAICallParams,
    OpenAICallResponse,
    OpenAICallResponseChunk,
    OpenAIDynamicConfig,
    OpenAIStream,
    OpenAITool,
)
from ..openai._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
)
from ._utils import setup_call

litellm_call = call_factory(
    TCallResponse=OpenAICallResponse,
    TCallResponseChunk=OpenAICallResponseChunk,
    TDynamicConfig=OpenAIDynamicConfig,
    TToolType=OpenAITool,
    TStream=OpenAIStream,
    TCallParams=OpenAICallParams,
    default_call_params=OpenAICallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,  # type: ignore
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the LiteLLM API with a typed function.

This decorator is used to wrap a typed function that calls the LiteLLM API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@litellm_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The LiteLLM model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the LiteLLM API call.
    **call_params: The `OpenAICallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an LiteLLM API call.
'''



"""Utility for converting `BaseMessageParam` to `ChatMessage`."""

from mistralai.models.chat_completion import ChatMessage

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ChatMessage],
) -> list[ChatMessage]:
    converted_message_params = []
    for message_param in message_params:
        if isinstance(message_param, ChatMessage):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(ChatMessage(**message_param.model_dump()))
        else:
            if len(content) != 1 or content[0].type != "text":
                raise ValueError("Mistral does not currently support multimodalities.")
            converted_message_params.append(
                ChatMessage(role=message_param.role, content=content[0].text)
            )
    return converted_message_params



"""This module contains the setup_call function for Cohere tools."""

import inspect
from typing import Any, Awaitable, Callable, cast

from cohere import AsyncClient, Client
from cohere.types import ChatMessage, NonStreamedChatResponse

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import CohereCallParams
from ..dynamic_config import CohereDynamicConfig
from ..tool import CohereTool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: Client | AsyncClient | None,
    fn: Callable[..., CohereDynamicConfig | Awaitable[CohereDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: CohereDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: CohereCallParams,
    extract: bool,
) -> tuple[
    Callable[..., NonStreamedChatResponse]
    | Callable[..., Awaitable[NonStreamedChatResponse]],
    str,
    list[ChatMessage],
    list[type[CohereTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, CohereTool, call_params
    )
    messages = cast(list[BaseMessageParam | ChatMessage], messages)
    messages = convert_message_params(messages)

    preamble = ""
    if "preamble" in call_kwargs and call_kwargs["preamble"] is not None:
        preamble += call_kwargs.pop("preamble")
    if messages[0].role == "SYSTEM":  # type: ignore
        if preamble:
            preamble += "\n\n"
        preamble += messages.pop(0).message
    if preamble:
        call_kwargs["preamble"] = preamble
    if len(messages) > 1:
        call_kwargs["chat_history"] = messages[:-1]
    if json_mode:
        # Cannot mutate ChatMessage in place
        messages[-1] = ChatMessage(
            role=messages[-1].role,  # type: ignore
            message=messages[-1].message
            + _utils.json_mode_content(tool_types[0] if tool_types else None),
            tool_calls=messages[-1].tool_calls,
        )
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
    call_kwargs |= {
        "model": model,
        "message": messages[-1].message,
    }

    if client is None:
        client = AsyncClient() if inspect.iscoroutinefunction(fn) else Client()

    def create_or_stream(stream: bool, **kwargs: Any):
        if stream:
            return client.chat_stream(**kwargs)
        return client.chat(**kwargs)

    return create_or_stream, prompt_template, messages, tool_types, call_kwargs  # type: ignore



"""Utility for converting `BaseMessageParam` to `ChatCompletionMessageParam`"""

from groq.types.chat import ChatCompletionMessageParam

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ChatCompletionMessageParam],
) -> list[ChatCompletionMessageParam]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif isinstance(content := message_param.content, str):
            converted_message_params.append(message_param.model_dump())
        else:
            if len(content) != 1 or content[0].type != "text":
                raise ValueError("Groq does not currently support multimodalities.")
            converted_message_params.append(
                {"role": message_param.role, "content": content[0].text}
            )
    return converted_message_params



"""The `BasePrompt` class for better prompt engineering."""

from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Iterable,
    ParamSpec,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ._stream import BaseStream
from ._utils import BaseType, format_template, get_metadata, get_prompt_template
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .metadata import Metadata

_P = ParamSpec("_P")
_R = TypeVar("_R")
_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)


class BasePrompt(BaseModel):
    """The base class for engineering prompts.

    This class is implemented as the base for all prompting needs. It is intended to
    work across various providers by providing a common prompt interface.

    Example:

    ```python
    from mirascope.core import BasePrompt, metadata

    @metadata({"tags": {"version:0001"}})
    class BookRecommendationPrompt(BasePrompt):
        prompt_template = "Recommend a {genre} book."

        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")

    prompt.messages()
    #> [{"role": "user", "content": "Please recommend a fantasy book."}]

    print(prompt)
    #> Please recommend a fantasy book.
    ```
    """

    def __str__(self) -> str:
        """Returns the formatted template."""
        return format_template(get_prompt_template(self), self.model_dump())

    def dump(self) -> dict[str, Any]:
        """Dumps the contents of the prompt into a dictionary."""
        return {
            "metadata": get_metadata(self, None),
            "prompt": str(self),
            "template": get_prompt_template(self),
            "inputs": self.model_dump(),
        }

    @overload
    def run(
        self,
        decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _BaseCallResponseT]
        ],
    ) -> _BaseCallResponseT: ...  # pragma: no cover

    @overload
    def run(
        self,
        decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _BaseStreamT]
        ],
    ) -> _BaseStreamT: ...  # pragma: no cover

    @overload
    def run(  # type: ignore
        self,
        decorator: Callable[
            [Callable[..., BaseDynamicConfig]], Callable[..., _ResponseModelT]
        ],
    ) -> _ResponseModelT: ...  # pragma: no cover

    @overload
    def run(
        self,
        decorator: Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., Iterable[_ResponseModelT]],
        ],
    ) -> Iterable[_ResponseModelT]: ...  # pragma: no cover

    def run(
        self,
        decorator: Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _BaseCallResponseT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _BaseStreamT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., _ResponseModelT],
        ]
        | Callable[
            [Callable[..., BaseDynamicConfig]],
            Callable[..., Iterable[_ResponseModelT]],
        ],
    ) -> (
        _BaseCallResponseT | _BaseStreamT | _ResponseModelT | Iterable[_ResponseModelT]
    ):
        """Returns the response of calling the API of the provided decorator."""
        kwargs = self.model_dump()
        args_str = ", ".join(kwargs.keys())
        namespace = {}
        exec(f"def fn({args_str}): ...", namespace)
        fn = namespace["fn"]
        return decorator(prompt_template(get_prompt_template(self))(fn))(**kwargs)

    @overload
    def run_async(
        self,
        decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseCallResponseT]],
        ],
    ) -> Awaitable[_BaseCallResponseT]: ...  # pragma: no cover

    @overload
    def run_async(
        self,
        decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseStreamT]],
        ],
    ) -> Awaitable[_BaseStreamT]: ...  # pragma: no cover

    @overload
    def run_async(
        self,
        decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_ResponseModelT]],
        ],
    ) -> Awaitable[_ResponseModelT]: ...  # pragma: no cover

    @overload
    def run_async(
        self,
        decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[AsyncIterable[_ResponseModelT]]],
        ],
    ) -> Awaitable[AsyncIterable[_ResponseModelT]]: ...  # pragma: no cover

    def run_async(
        self,
        decorator: Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseCallResponseT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_BaseStreamT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[_ResponseModelT]],
        ]
        | Callable[
            [Callable[..., Awaitable[BaseDynamicConfig]]],
            Callable[..., Awaitable[AsyncIterable[_ResponseModelT]]],
        ],
    ) -> (
        Awaitable[_BaseCallResponseT]
        | Awaitable[_BaseStreamT]
        | Awaitable[_ResponseModelT]
        | Awaitable[AsyncIterable[_ResponseModelT]]
    ):
        kwargs = self.model_dump()
        args_str = ", ".join(kwargs.keys())
        namespace = {}
        exec(f"async def fn({args_str}): ...", namespace)
        fn = namespace["fn"]
        return decorator(prompt_template(get_prompt_template(self))(fn))(**kwargs)


_BasePromptT = TypeVar("_BasePromptT", bound=BasePrompt)


def prompt_template(template: str):
    """A decorator for setting the `prompt_template` of a `BasePrompt` or `call`."""

    @overload
    def inner(prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...  # pragma: no cover

    @overload
    def inner(prompt: Callable[_P, _R]) -> Callable[_P, _R]: ...  # pragma: no cover

    def inner(
        prompt: type[_BasePromptT] | Callable[_P, _R],
    ) -> type[_BasePromptT] | Callable[_P, _R]:
        setattr(prompt, "_prompt_template", template)
        return prompt

    return inner


def metadata(metadata: Metadata):
    """A decorator for adding metadata to a `BasePrompt` or `call`.

    Adding this decorator to a `BasePrompt` or `call` updates the `metadata` annotation
    to the given value. This is useful for adding metadata to a `BasePrompt` or `call`
    that can be used for logging or filtering.

    Example:

    ```python
    from mirascope.core import BasePrompt, metadata

    @metadata({"tags": {"version:0001", "books"}})
    class BookRecommendationPrompt(BasePrompt):
        prompt_template = "Recommend a {genre} book."

        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")

    print(prompt.dump()["metadata"])
    #> {"metadata": {"version:0001", "books"}}
    ```

    Returns:
        Decorator function that updates the `metadata` attribute of the decorated input.
    """

    @overload
    def inner(prompt: type[_BasePromptT]) -> type[_BasePromptT]: ...  # pragma: no cover

    @overload
    def inner(prompt: Callable[_P, _R]) -> Callable[_P, _R]: ...  # pragma: no cover

    def inner(
        prompt: type[_BasePromptT] | Callable[_P, _R],
    ) -> type[_BasePromptT] | Callable[_P, _R]:
        """Updates the `metadata` class attribute to the given value."""
        setattr(prompt, "_metadata", metadata)
        return prompt

    return inner



"""The `OpenAIStream` class for convenience around streaming LLM calls."""

from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessage,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCall,
    ChatCompletionMessageToolCallParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.chat_completion import Choice
from openai.types.chat.chat_completion_message_tool_call_param import Function
from openai.types.completion_usage import CompletionUsage

from ..base._stream import BaseStream
from ._utils import calculate_cost
from .call_params import OpenAICallParams
from .call_response import OpenAICallResponse
from .call_response_chunk import OpenAICallResponseChunk
from .dynamic_config import OpenAIDynamicConfig
from .tool import OpenAITool

FinishReason = Choice.__annotations__["finish_reason"]


class OpenAIStream(
    BaseStream[
        OpenAICallResponse,
        OpenAICallResponseChunk,
        ChatCompletionUserMessageParam,
        ChatCompletionAssistantMessageParam,
        ChatCompletionToolMessageParam,
        ChatCompletionMessageParam,
        OpenAITool,
        OpenAIDynamicConfig,
        OpenAICallParams,
        FinishReason,
    ]
):
    _provider = "openai"

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    def _construct_message_param(
        self,
        tool_calls: list[ChatCompletionMessageToolCall] | None = None,
        content: str | None = None,
    ) -> ChatCompletionAssistantMessageParam:
        """Constructs the message parameter for the assistant."""
        message_param = ChatCompletionAssistantMessageParam(
            role="assistant", content=content
        )
        if tool_calls:
            message_param["tool_calls"] = [
                ChatCompletionMessageToolCallParam(
                    type="function",
                    function=Function(
                        arguments=tool_call.function.arguments,
                        name=tool_call.function.name,
                    ),
                    id=tool_call.id,
                )
                for tool_call in tool_calls
            ]
        return message_param

    def construct_call_response(self) -> OpenAICallResponse:
        """Constructs the call response from a consumed OpenAIStream."""
        if not hasattr(self, "message_param"):
            raise ValueError(
                "No stream response, check if the stream has been consumed."
            )
        message = {
            "role": self.message_param["role"],
            "content": self.message_param.get("content", ""),
            "tool_calls": self.message_param.get("tool_calls", []),
        }
        if not self.input_tokens and not self.output_tokens:
            usage = None
        else:
            usage = CompletionUsage(
                prompt_tokens=int(self.input_tokens or 0),
                completion_tokens=int(self.output_tokens or 0),
                total_tokens=int(self.input_tokens or 0) + int(self.output_tokens or 0),
            )
        completion = ChatCompletion(
            id=self.id if self.id else "",
            model=self.model,
            choices=[
                Choice(
                    finish_reason=self.finish_reasons[0]
                    if self.finish_reasons
                    else "stop",
                    index=0,
                    message=ChatCompletionMessage.model_validate(message),
                )
            ],
            created=0,
            object="chat.completion",
            usage=usage,
        )
        return OpenAICallResponse(
            metadata=self.metadata,
            response=completion,
            tool_types=self.tool_types,
            prompt_template=self.prompt_template,
            fn_args=self.fn_args if self.fn_args else {},
            dynamic_config=self.dynamic_config,
            messages=self.messages,
            call_params=self.call_params,
            call_kwargs=self.call_kwargs,
            user_message_param=self.user_message_param,
            start_time=self.start_time,
            end_time=self.end_time,
        )



"""Utility for setting up a provider-specific call."""

import inspect
from typing import Any, Awaitable, Callable, TypeVar

from ..call_params import BaseCallParams
from ..dynamic_config import BaseDynamicConfig
from ..message_param import BaseMessageParam
from ..tool import BaseTool
from ._convert_base_model_to_base_tool import convert_base_model_to_base_tool
from ._convert_function_to_base_tool import convert_function_to_base_tool
from ._get_prompt_template import get_prompt_template
from ._parse_prompt_messages import parse_prompt_messages

_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)


def setup_call(
    fn: Callable[..., _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
    fn_args: dict[str, Any],
    dynamic_config: _BaseDynamicConfigT,
    tools: list[type[BaseTool] | Callable] | None,
    tool_type: type[_BaseToolT],
    call_params: BaseCallParams,
) -> tuple[
    str,
    list[BaseMessageParam | Any],
    list[type[_BaseToolT]] | None,
    dict[str, Any],
]:
    call_kwargs = call_params.copy()
    prompt_template, messages, computed_fields = None, None, None
    if dynamic_config is not None:
        computed_fields = dynamic_config.get("computed_fields", None)
        tools = dynamic_config.get("tools", tools)
        messages = dynamic_config.get("messages", None)
        dynamic_call_params = dynamic_config.get("call_params", None)
        if dynamic_call_params:
            call_kwargs |= dynamic_call_params

    if not messages:
        prompt_template = get_prompt_template(fn)
        assert prompt_template is not None, "The function must have a docstring."
        if computed_fields:
            fn_args |= computed_fields
        messages = parse_prompt_messages(
            roles=["system", "user", "assistant"],
            template=prompt_template,
            attrs=fn_args,
        )

    tool_types = None
    if tools:
        tool_types = [
            convert_base_model_to_base_tool(tool, tool_type)
            if inspect.isclass(tool)
            else convert_function_to_base_tool(tool, tool_type)
            for tool in tools
        ]
        call_kwargs["tools"] = [tool_type.tool_schema() for tool_type in tool_types]  # type: ignore

    return prompt_template, messages, tool_types, call_kwargs  # type: ignore



"""This module contains the `OpenAICallResponse` class."""

from openai.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.completion_usage import CompletionUsage
from pydantic import SkipValidation, computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import OpenAICallParams
from .dynamic_config import OpenAIDynamicConfig
from .tool import OpenAITool


class OpenAICallResponse(
    BaseCallResponse[
        ChatCompletion,
        OpenAITool,
        OpenAIDynamicConfig,
        ChatCompletionMessageParam,
        OpenAICallParams,
        ChatCompletionUserMessageParam,
    ]
):
    '''A convenience wrapper around the OpenAI `ChatCompletion` response.

    When calling the OpenAI API using a function decorated with `openai_call`, the
    response will be an `OpenAICallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.openai import openai_call

    @openai_call(model="gpt-4o")
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    response = recommend_book("fantasy")  # response is an `OpenAICallResponse` instance
    print(response.content)
    #> Sure! I would recommend...
    ```
    '''

    response: SkipValidation[ChatCompletion]

    _provider = "openai"

    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        message = self.response.choices[0].message
        return message.content if message.content is not None else ""

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(choice.finish_reason) for choice in self.response.choices]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> CompletionUsage | None:
        """Returns the usage of the chat completion."""
        return self.response.usage

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens if self.usage else None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens if self.usage else None

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> ChatCompletionAssistantMessageParam:
        """Returns the assistants's response as a message parameter."""
        return self.response.choices[0].message.model_dump(exclude={"function_call"})  # type: ignore

    @computed_field
    @property
    def tools(self) -> list[OpenAITool] | None:
        """Returns any available tool calls as their `OpenAITool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        tool_calls = self.response.choices[0].message.tool_calls
        if not self.tool_types or not tool_calls:
            return None

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> OpenAITool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[OpenAITool, str]]
    ) -> list[ChatCompletionToolMessageParam]:
        """Returns the tool message parameters for tool call results."""
        return [
            ChatCompletionToolMessageParam(
                role="tool",
                content=output,
                tool_call_id=tool.tool_call.id,
                name=tool._name(),  # type: ignore
            )
            for tool, output in tools_and_outputs
        ]



"""The parameters to use when calling the Mistral API."""

from __future__ import annotations

from mistralai.models.chat_completion import ResponseFormat, ToolChoice
from typing_extensions import NotRequired

from ..base import BaseCallParams


class MistralCallParams(BaseCallParams):
    """The parameters to use when calling the Mistral API."""

    endpoint: NotRequired[str | None]
    max_tokens: NotRequired[int | None]
    random_seed: NotRequired[int | None]
    response_format: NotRequired[ResponseFormat | None]
    safe_mode: NotRequired[bool | None]
    safe_prompt: NotRequired[bool | None]
    temperature: NotRequired[float | None]
    tool_choice: NotRequired[ToolChoice | None]
    top_p: NotRequired[float | None]



"""This module contains the setup_call function for OpenAI tools."""

import inspect
from typing import Any, Awaitable, Callable

from litellm import acompletion, completion
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam

from ...base import BaseTool
from ...openai import OpenAICallParams, OpenAIDynamicConfig, OpenAITool
from ...openai._utils import setup_call as setup_call_openai


def setup_call(
    *,
    model: str,
    client: OpenAI | AsyncOpenAI | AzureOpenAI | AsyncAzureOpenAI | None,
    fn: Callable[..., OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion] | Callable[..., Awaitable[ChatCompletion]],
    str,
    list[ChatCompletionMessageParam],
    list[type[OpenAITool]] | None,
    dict[str, Any],
]:
    _, prompt_template, messages, tool_types, call_kwargs = setup_call_openai(
        model=model,
        client=client,
        fn=fn,
        fn_args=fn_args,
        dynamic_config=dynamic_config,
        tools=tools,
        json_mode=json_mode,
        call_params=call_params,
        extract=extract,
    )
    create = acompletion if inspect.iscoroutinefunction(fn) else completion
    return create, prompt_template, messages, tool_types, call_kwargs  # type: ignore



"""Mirascope package."""

import importlib.metadata
from contextlib import suppress

with suppress(ImportError):
    from . import core as core

with suppress(ImportError):
    from . import integrations as integrations

__version__ = importlib.metadata.version("mirascope")

__all__ = ["core", "integrations", "__version__"]



"""This module provides a function to parse messages from a prompt template."""

import re
from typing import Any, TypeVar

from pydantic import BaseModel

from ..message_param import BaseMessageParam
from ._get_template_variables import get_template_variables
from ._parse_content_template import parse_content_template

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def parse_prompt_messages(
    roles: list[str], template: str, attrs: dict[str, Any]
) -> list[BaseMessageParam | Any]:
    """Returns messages parsed from the provided prompt `template`.

    Raises:
        ValueError: if `MESSAGES` keyword is used with a non-list attribute.
    """
    messages = []
    re_roles = "|".join([role.upper() for role in roles] + ["MESSAGES"])
    for match in re.finditer(rf"({re_roles}):((.|\n)+?)(?=({re_roles}):|\Z)", template):
        role, content_template = match.group(1).lower(), match.group(2).strip()
        if role == "messages":
            template_variables = get_template_variables(content_template, False)
            if template_variables[0].startswith("self"):
                if "self" not in attrs:
                    raise ValueError(
                        "MESSAGES keyword used with `self.` but `self` was not found."
                    )
                attr = getattr(attrs["self"], template_variables[0][5:])
            else:
                attr = attrs[template_variables[0]]
            if attr is None or not isinstance(attr, list):
                raise ValueError(
                    f"MESSAGES keyword used with attribute `{template_variables[0]}`"
                    ", which is not a `list` of messages."
                )
            messages += attr
        else:
            content = parse_content_template(role, content_template, attrs)
            if content:
                messages.append(content)
    if len(messages) == 0:
        content = parse_content_template("user", template, attrs)
        if content:
            messages.append(content)
    return messages



"""The `OpenAITool` class for easy tool usage with OpenAI LLM calls."""

from __future__ import annotations

import copy

from anthropic.types import ToolParam, ToolUseBlock
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class AnthropicTool(BaseTool):
    """A class for defining tools for Anthropic LLM calls."""

    tool_call: SkipJsonSchema[ToolUseBlock]

    @classmethod
    def tool_schema(cls) -> ToolParam:
        """Constructs a `ToolParam` tool schema from the `BaseModel` schema defined."""
        return ToolParam(
            input_schema=cls.model_tool_schema(),
            name=cls._name(),
            description=cls._description(),
        )

    @classmethod
    def from_tool_call(cls, tool_call: ToolUseBlock) -> AnthropicTool:
        """Constructs an `AnthropicTool` instance from a `tool_call`."""
        model_json = copy.deepcopy(tool_call.input)
        model_json["tool_call"] = tool_call.model_dump()  # type: ignore
        return cls.model_validate(model_json)



"""This module contains the setup_call function, which is used to set up the"""

import inspect
from typing import Any, Awaitable, Callable, cast

from google.generativeai import GenerativeModel  # type: ignore
from google.generativeai.types import (  # type: ignore
    AsyncGenerateContentResponse,
    ContentDict,
    GenerateContentResponse,
)

from ...base import BaseMessageParam, BaseTool, _utils
from ..call_params import GeminiCallParams
from ..dynamic_config import GeminiDynamicConfig
from ..tool import GeminiTool
from ._convert_message_params import convert_message_params


def setup_call(
    *,
    model: str,
    client: GenerativeModel | None,
    fn: Callable[..., GeminiDynamicConfig | Awaitable[GeminiDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GeminiDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GeminiCallParams,
    extract: bool = False,
) -> tuple[
    Callable[..., GenerateContentResponse]
    | Callable[..., Awaitable[AsyncGenerateContentResponse]],
    str,
    list[ContentDict],
    list[type[GeminiTool]] | None,
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, GeminiTool, call_params
    )
    messages = cast(list[BaseMessageParam | ContentDict], messages)
    messages = convert_message_params(messages)
    if json_mode:
        generation_config = call_kwargs.get("generation_config", {})
        generation_config["response_mime_type"] = "application/json"
        call_kwargs["generation_config"] = generation_config
        messages[-1]["parts"].append(
            _utils.json_mode_content(tool_types[0] if tool_types else None)
        )
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        tool_config = call_kwargs.get("tool_config", {})
        tool_config["function_calling_config"] = {"mode": "auto"}
        call_kwargs["tool_config"] = tool_config
    call_kwargs |= {"contents": messages}

    if client is None:
        client = GenerativeModel(model_name=model)
    create = (
        client.generate_content_async
        if inspect.iscoroutinefunction(fn)
        else client.generate_content
    )

    return create, prompt_template, messages, tool_types, call_kwargs



"""Beta implementation of the new OpenAI parse functionality."""

import inspect
from functools import wraps
from typing import Awaitable, Callable, Literal, ParamSpec, TypeVar, cast, overload

from pydantic import BaseModel

from mirascope.core.base import BaseMessageParam, _utils
from mirascope.core.openai import OpenAICallParams, OpenAIDynamicConfig, OpenAITool
from mirascope.core.openai._utils import convert_message_params
from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletionMessageParam

_ResponseModelT = TypeVar("_ResponseModelT", bound=_utils.BaseType | BaseModel)
_P = ParamSpec("_P")


def parse(
    model: Literal["gpt-4o-2024-08-06"],
    response_model: type[_ResponseModelT],
    call_params: OpenAICallParams | None = None,
):
    """Beta implementation of the new OpenAI parse functionality."""

    @overload
    def decorator(
        fn: Callable[_P, OpenAIDynamicConfig],
    ) -> Callable[_P, _ResponseModelT]: ...

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[OpenAIDynamicConfig]],
    ) -> Callable[_P, Awaitable[_ResponseModelT]]: ...

    def decorator(
        fn: Callable[_P, OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    ) -> Callable[_P, _ResponseModelT | Awaitable[_ResponseModelT]]:
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert _utils.SetupCall.fn_is_sync(fn)
            fn_args = _utils.get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            _, messages, _, call_kwargs = _utils.setup_call(
                fn,
                fn_args,
                dynamic_config,
                None,
                OpenAITool,
                call_params or OpenAICallParams(),
            )
            messages = cast(
                list[BaseMessageParam | ChatCompletionMessageParam], messages
            )
            messages = convert_message_params(messages)
            messages.append({"role": "user", "content": _utils.json_mode_content(None)})
            client = OpenAI()
            call_kwargs.pop("response_format", None)
            completion = client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                response_format=response_model,
                **call_kwargs,
            )
            message = completion.choices[0].message
            if message.parsed:
                return message.parsed
            raise RuntimeError(f"{message.refusal}")

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert _utils.SetupCall.fn_is_async(fn)
            fn_args = _utils.get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            _, messages, _, call_kwargs = _utils.setup_call(
                fn,
                fn_args,
                dynamic_config,
                None,
                OpenAITool,
                call_params or OpenAICallParams(),
            )
            messages = cast(
                list[BaseMessageParam | ChatCompletionMessageParam], messages
            )
            messages = convert_message_params(messages)
            messages.append({"role": "user", "content": _utils.json_mode_content(None)})
            client = AsyncOpenAI()
            call_kwargs.pop("response_format", None)
            completion = await client.beta.chat.completions.parse(
                model=model,
                messages=messages,
                response_format=response_model,
                **call_kwargs,
            )
            message = completion.choices[0].message
            if message.parsed:
                return message.parsed
            raise RuntimeError(f"{message.refusal}")

        return inner_async if is_async else inner

    return decorator



"""Types for interacting with Weaviate using Mirascope."""

from typing import Any, Literal, Optional

import weaviate.types as wt
from pydantic import BaseModel, ConfigDict

from ..base.document import Document
from ..base.vectorstore_params import BaseVectorStoreParams


class WeaviateSettings(BaseModel):
    mode: Literal["local", "cloud", "custom", "embedded"] = "local"
    headers: Optional[dict[str, str]] = None
    cluster_url: Optional[str] = None
    auth_credentials: Optional[Any] = None
    hostname: Optional[str] = None
    port: Optional[int] = None
    grpc_port: Optional[int] = None
    additional_config: Optional[Any] = None
    version: Optional[str] = None
    persistence_data_path: Optional[str] = None
    binary_path: Optional[str] = None
    environment_variables: Optional[dict[str, str]] = None
    http_host: Optional[str] = None
    http_port: Optional[int] = None
    http_secure: Optional[bool] = None
    grpc_host: Optional[str] = None
    grpc_secure: Optional[bool] = None

    skip_init_checks: Optional[bool] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def kwargs(self) -> dict[str, Any]:
        kwargs = {}
        exclude = {"mode"}
        if self.auth_credentials:
            kwargs["auth_credentials"] = self.auth_credentials
            exclude.add("auth_credentials")
        if self.additional_config:
            kwargs["additional_config"] = self.additional_config
            exclude.add("additional_config")
        additional_kwargs = {
            key: value
            for key, value in self.model_dump(exclude=exclude).items()
            if value is not None
        }
        kwargs.update(additional_kwargs)
        return kwargs


class WeaviateParams(BaseVectorStoreParams):
    vectorizer_config: Optional[Any] = None
    generative_config: Optional[Any] = None

    def kwargs(self) -> dict[str, Any]:
        kwargs = {}
        exclude = set()
        if self.vectorizer_config:
            kwargs["vectorizer_config"] = self.vectorizer_config
            exclude = {"vectorizer_config"}
        additional_kwargs = {
            key: value
            for key, value in self.model_dump(exclude=exclude).items()
            if value is not None
        }
        kwargs.update(additional_kwargs)
        return kwargs


class WeaviateQueryResult(BaseModel):
    id: wt.UUID
    documents: Optional[list[Document]]
    metadata: dict[str, Any]
    collection: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def from_response(response_object):
        model_dict = {}
        id = str(response_object.uuid)
        text = response_object.properties["text"]
        model_dict["id"] = id
        doc = Document(id=id, text=text)
        model_dict["documents"] = [doc]
        model_dict["metadata"] = response_object.metadata.__dict__
        model_dict["collection"] = response_object.collection

        return WeaviateQueryResult.model_validate(model_dict)



"""This module contains the `GeminiCallResponseChunk` class."""

from google.ai.generativelanguage import Candidate
from google.generativeai.types import (  # type: ignore
    GenerateContentResponse,
)

from ..base import BaseCallResponseChunk


class GeminiCallResponseChunk(
    BaseCallResponseChunk[GenerateContentResponse, Candidate.FinishReason]
):
    """Convenience wrapper around chat completion streaming chunks.

    When using Mirascope's convenience wrappers to interact with Gemini models via
    `GeminiCall`, responses using `GeminiCall.stream()` will return a
    `GeminiCallResponseChunk`, whereby the implemented properties allow for simpler
    syntax and a convenient developer experience.

    Example:

    ```python
    from mirascope.gemini import GeminiCall


    class Math(GeminiCall):
        prompt_template = "What is 1 + 2?"


    content = ""
    for chunk in Math().stream():
        content += chunk.content
        print(content)
    #> 1
    #  1 +
    #  1 + 2
    #  1 + 2 equals
    #  1 + 2 equals
    #  1 + 2 equals 3
    #  1 + 2 equals 3.
    ```
    """

    @property
    def content(self) -> str:
        """Returns the chunk content for the 0th choice."""
        return self.chunk.candidates[0].content.parts[0].text

    @property
    def finish_reasons(self) -> list[Candidate.FinishReason]:
        """Returns the finish reasons of the response."""
        return [candidate.finish_reason for candidate in self.chunk.candidates]

    @property
    def model(self) -> None:
        """Returns the model name.

        google.generativeai does not return model, so we return None
        """
        return None

    @property
    def id(self) -> str | None:
        """Returns the id of the response.

        google.generativeai does not return an id
        """
        return None

    @property
    def usage(self) -> None:
        """Returns the usage of the chat completion.

        google.generativeai does not have Usage, so we return None
        """
        return None

    @property
    def input_tokens(self) -> None:
        """Returns the number of input tokens."""
        return None

    @property
    def output_tokens(self) -> None:
        """Returns the number of output tokens."""
        return None



"""Utility for pulling the `prompt_template` from a call or `BasePrompt`."""

import inspect
import os
from typing import Callable

from pydantic import BaseModel


def get_prompt_template(fn: Callable | BaseModel) -> str:
    """Get the metadata from the function and merge with any dynamic metadata."""
    prompt_template = getattr(fn, "_prompt_template", None)
    if prompt_template:
        return prompt_template

    docstring_prompt_enabled = os.getenv("MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE")
    doc = fn.__doc__
    if not doc:
        raise ValueError("No prompt template set!")
    if docstring_prompt_enabled != "ENABLED":
        raise ValueError("You must explicitly enable docstring prompt templates.")
    return inspect.cleandoc(doc)



"""Utility for getting the possible most recent user message."""

from typing import TypeVar

_T = TypeVar("_T")


def get_possible_user_message_param(messages: list[_T]) -> _T | None:
    """Get the possible most recent user message."""
    if not messages:
        return None
    most_recent_message = messages[-1]
    if (
        isinstance(most_recent_message, dict)
        and "role" in most_recent_message
        and most_recent_message["role"] == "user"
    ):
        return most_recent_message
    if (
        hasattr(most_recent_message, "role")
        and getattr(most_recent_message, "role") == "user"
    ):
        return most_recent_message
    return None



"""The `create_factory` method for generating provider specific create decorators."""

import datetime
import inspect
from functools import wraps
from typing import Awaitable, Callable, ParamSpec, TypeVar, overload

from ._utils import (
    SetupCall,
    get_fn_args,
    get_metadata,
    get_possible_user_message_param,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_P = ParamSpec("_P")


def create_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
):
    """Returns the wrapped function with the provider specific interfaces."""

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _BaseCallResponseT | _ParsedOutputT]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        _BaseCallResponseT
        | _ParsedOutputT
        | Awaitable[_BaseCallResponseT | _ParsedOutputT],
    ]:
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> TCallResponse | _ParsedOutputT:
            assert SetupCall.fn_is_sync(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )
            start_time = datetime.datetime.now().timestamp() * 1000
            response = create(stream=False, **call_kwargs)
            end_time = datetime.datetime.now().timestamp() * 1000
            output = TCallResponse(
                metadata=get_metadata(fn, dynamic_config),
                response=response,
                tool_types=tool_types,  # type: ignore
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                call_kwargs=call_kwargs,
                user_message_param=get_possible_user_message_param(messages),
                start_time=start_time,
                end_time=end_time,
            )
            output._model = model
            return output if not output_parser else output_parser(output)

        @wraps(fn)
        async def inner_async(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> TCallResponse | _ParsedOutputT:
            assert SetupCall.fn_is_async(fn)
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )
            start_time = datetime.datetime.now().timestamp() * 1000
            response = await create(stream=False, **call_kwargs)
            end_time = datetime.datetime.now().timestamp() * 1000
            output = TCallResponse(
                metadata=get_metadata(fn, dynamic_config),
                response=response,
                tool_types=tool_types,  # type: ignore
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                call_kwargs=call_kwargs,
                user_message_param=get_possible_user_message_param(messages),
                start_time=start_time,
                end_time=end_time,
            )
            output._model = model
            return output if not output_parser else output_parser(output)

        return inner_async if is_async else inner

    return decorator



"""The `openai_call` decorator for functions as LLM calls."""

from ..base import call_factory
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    setup_call,
)
from .call_params import OpenAICallParams
from .call_response import OpenAICallResponse
from .call_response_chunk import OpenAICallResponseChunk
from .dynamic_config import OpenAIDynamicConfig
from .stream import OpenAIStream
from .tool import OpenAITool

openai_call = call_factory(
    TCallResponse=OpenAICallResponse,
    TCallResponseChunk=OpenAICallResponseChunk,
    TDynamicConfig=OpenAIDynamicConfig,
    TToolType=OpenAITool,
    TStream=OpenAIStream,
    TCallParams=OpenAICallParams,
    default_call_params=OpenAICallParams(),
    setup_call=setup_call,  # type: ignore
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
)
'''A decorator for calling the OpenAI API with a typed function.

This decorator is used to wrap a typed function that calls the OpenAI API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@openai_call(model="gpt-4o")
def recommend_book(genre: str):
    """Recommend a {genre} book."""

response = recommend_book("fantasy")
```

Args:
    model: The OpenAI model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the OpenAI API call.
    **call_params: The `OpenAICallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an OpenAI API call.
'''



"""The parameters to use when calling the Cohere API."""

from __future__ import annotations

from typing import Sequence

from cohere.types import (
    ChatConnector,
    ChatDocument,
    ChatMessage,
    ChatRequestPromptTruncation,
    ToolResult,
)
from typing_extensions import NotRequired

from ..base import BaseCallParams


class CohereCallParams(BaseCallParams):
    """The parameters to use when calling the Cohere API."""

    preamble: NotRequired[str | None]
    chat_history: NotRequired[Sequence[ChatMessage] | None]
    conversation_id: NotRequired[str | None]
    prompt_truncation: NotRequired[ChatRequestPromptTruncation | None]
    connectors: NotRequired[Sequence[ChatConnector] | None]
    search_queries_only: NotRequired[bool | None]
    documents: NotRequired[Sequence[ChatDocument] | None]
    temperature: NotRequired[float | None]
    max_tokens: NotRequired[int | None]
    max_input_tokens: NotRequired[int | None]
    k: NotRequired[int | None]
    p: NotRequired[float | None]
    seed: NotRequired[int | None]
    stop_sequences: NotRequired[Sequence[str] | None]
    frequency_penalty: NotRequired[float | None]
    presence_penalty: NotRequired[float | None]
    raw_prompting: NotRequired[bool | None]
    tool_results: NotRequired[Sequence[ToolResult] | None]



"""Get JSON output from a Gemini response."""

import json

from proto.marshal.collections import RepeatedComposite

from ..call_response import GeminiCallResponse
from ..call_response_chunk import GeminiCallResponseChunk


def get_json_output(
    response: GeminiCallResponse | GeminiCallResponseChunk, json_mode: bool
) -> str:
    """Extracts the JSON output from a Gemini response."""
    if isinstance(response, GeminiCallResponse):
        if json_mode and response.content:
            return response.content
        elif tool_calls := [
            part.function_call
            for part in response.response.parts
            if part.function_call.args
        ]:
            return json.dumps(
                {
                    k: v
                    if not isinstance(v, RepeatedComposite)
                    else [item for item in v]
                    for k, v in tool_calls[0].args.items()
                }
            )
        else:
            raise ValueError("No tool call or JSON object found in response.")
    elif not json_mode:
        raise ValueError("Gemini only supports structured streaming in json mode.")
    return response.content



"""The `MistralTool` class for easy tool usage with Mistral LLM calls."""

from __future__ import annotations

from typing import Any

import jiter
from mistralai.models.chat_completion import ToolCall
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class MistralTool(BaseTool):
    """A class for defining tools for Mistral LLM calls."""

    tool_call: SkipJsonSchema[ToolCall]

    @classmethod
    def tool_schema(cls) -> dict[str, Any]:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined."""
        fn: dict[str, Any] = {"name": cls._name(), "description": cls._description()}
        model_schema = cls.model_tool_schema()
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        return {"function": fn, "type": "function"}

    @classmethod
    def from_tool_call(cls, tool_call: ToolCall) -> MistralTool:
        """Constructs an `MistralTool` instance from a `tool_call`."""
        model_json = jiter.from_json(tool_call.function.arguments.encode())
        model_json["tool_call"] = tool_call.model_dump()
        return cls.model_validate(model_json)



"""Handles the stream of completion chunks."""

from collections.abc import AsyncGenerator, Generator

from groq.types.chat import ChatCompletionChunk, ChatCompletionMessageToolCall
from groq.types.chat.chat_completion_message_tool_call import Function

from ..call_response_chunk import GroqCallResponseChunk
from ..tool import GroqTool


def _handle_chunk(
    chunk: ChatCompletionChunk,
    current_tool_call: ChatCompletionMessageToolCall,
    current_tool_type: type[GroqTool] | None,
    tool_types: list[type[GroqTool]] | None,
) -> tuple[
    GroqTool | None,
    ChatCompletionMessageToolCall,
    type[GroqTool] | None,
]:
    """Handles a chunk of the stream."""
    if not tool_types or not (tool_calls := chunk.choices[0].delta.tool_calls):
        return None, current_tool_call, current_tool_type

    tool_call = tool_calls[0]
    # Reset on new tool
    if tool_call.id and tool_call.function is not None:
        previous_tool_call = current_tool_call.model_copy()
        previous_tool_type = current_tool_type
        current_tool_call = ChatCompletionMessageToolCall(
            id=tool_call.id,
            function=Function(
                arguments="",
                name=tool_call.function.name if tool_call.function.name else "",
            ),
            type="function",
        )
        current_tool_type = None
        for tool_type in tool_types:
            if tool_type._name() == tool_call.function.name:
                current_tool_type = tool_type
                break
        if current_tool_type is None:
            raise RuntimeError(
                f"Unknown tool type in stream: {tool_call.function.name}"
            )  # pragma: no cover
        if (
            previous_tool_call.id
            and previous_tool_call.function.arguments
            and previous_tool_type is not None
        ):
            return (
                previous_tool_type.from_tool_call(previous_tool_call),
                current_tool_call,
                current_tool_type,
            )

    # Update arguments with each chunk
    if tool_call.function and tool_call.function.arguments:
        current_tool_call.function.arguments += tool_call.function.arguments

    return None, current_tool_call, current_tool_type


def handle_stream(
    stream: Generator[ChatCompletionChunk, None, None],
    tool_types: list[type[GroqTool]] | None,
) -> Generator[tuple[GroqCallResponseChunk, GroqTool | None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed."""
    current_tool_call = ChatCompletionMessageToolCall(
        id="", function=Function(arguments="", name=""), type="function"
    )
    current_tool_type = None
    for chunk in stream:
        if not tool_types or not chunk.choices[0].delta.tool_calls:
            if current_tool_type:
                yield (
                    GroqCallResponseChunk(chunk=chunk),
                    current_tool_type.from_tool_call(current_tool_call),
                )
                current_tool_type = None
            else:
                yield GroqCallResponseChunk(chunk=chunk), None
        tool, current_tool_call, current_tool_type = _handle_chunk(
            chunk,
            current_tool_call,
            current_tool_type,
            tool_types,
        )
        if tool is not None:
            yield GroqCallResponseChunk(chunk=chunk), tool


async def handle_stream_async(
    stream: AsyncGenerator[ChatCompletionChunk, None],
    tool_types: list[type[GroqTool]] | None,
) -> AsyncGenerator[tuple[GroqCallResponseChunk, GroqTool | None], None]:
    """Async iterator over the stream and constructs tools as they are streamed."""
    current_tool_call = ChatCompletionMessageToolCall(
        id="", function=Function(arguments="", name=""), type="function"
    )
    current_tool_type = None
    async for chunk in stream:
        if not tool_types or not chunk.choices[0].delta.tool_calls:
            if current_tool_type:
                yield (
                    GroqCallResponseChunk(chunk=chunk),
                    current_tool_type.from_tool_call(current_tool_call),
                )
                current_tool_type = None
            else:
                yield GroqCallResponseChunk(chunk=chunk), None
        tool, current_tool_call, current_tool_type = _handle_chunk(
            chunk,
            current_tool_call,
            current_tool_type,
            tool_types,
        )
        if tool is not None:
            yield GroqCallResponseChunk(chunk=chunk), tool



"""Mirascope x Langfuse Integration."""

from typing import (
    Awaitable,
    Callable,
    ParamSpec,
    TypeVar,
)

from langfuse.decorators import observe
from pydantic import BaseModel

from mirascope.core.base._structured_stream import BaseStructuredStream

from ...core.base import BaseCallResponse
from ...core.base._stream import BaseStream
from ..middleware_factory import middleware_decorator
from ._utils import (
    handle_call_response,
    handle_call_response_async,
    handle_response_model,
    handle_response_model_async,
    handle_stream,
    handle_stream_async,
    handle_structured_stream,
    handle_structured_stream_async,
)

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_BaseModelT = TypeVar("_BaseModelT", bound=BaseModel)
_BaseStructuredStreamT = TypeVar("_BaseStructuredStreamT", bound=BaseStructuredStream)
_P = ParamSpec("_P")
SyncFunc = Callable[
    _P, _BaseCallResponseT | _BaseStreamT | _BaseModelT | _BaseStructuredStreamT
]
AsyncFunc = Callable[
    _P,
    Awaitable[_BaseCallResponseT | _BaseStreamT | _BaseModelT | _BaseStructuredStreamT],
]


def with_langfuse(fn):
    """Wraps a Mirascope function with Langfuse."""

    return middleware_decorator(
        fn,
        custom_decorator=observe(
            name=fn.__name__,
            as_type="generation",
            capture_input=False,
            capture_output=False,
        ),
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_response_model=handle_response_model,
        handle_response_model_async=handle_response_model_async,
        handle_structured_stream=handle_structured_stream,
        handle_structured_stream_async=handle_structured_stream_async,
    )



"""This module defines the base class for structured streams."""

import inspect
from collections.abc import AsyncGenerator, Generator
from functools import wraps
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    ParamSpec,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ._stream import BaseStream, stream_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    SetupCall,
    extract_tool_return,
    setup_extract_tool,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)


class BaseStructuredStream(Generic[_ResponseModelT]):
    """A base class for streaming structured outputs from LLMs."""

    stream: BaseStream
    response_model: type[_ResponseModelT]
    constructed_response_model: _ResponseModelT

    def __init__(
        self,
        *,
        stream: BaseStream,
        response_model: type[_ResponseModelT],
    ):
        """Initializes an instance of `BaseStructuredStream`."""
        self.stream = stream
        self.response_model = response_model

    def __iter__(self) -> Generator[_ResponseModelT, None, None]:
        """Iterates over the stream and extracts structured outputs."""
        json_output = ""
        for chunk, _ in self.stream:
            json_output += chunk.content
            if json_output and json_output[0] != "{":
                try:
                    json_start = json_output.index("{")
                    json_output = json_output[json_start:]
                except ValueError:
                    json_output = ""
            if chunk.model is not None:
                self.stream.model = chunk.model
            if json_output:
                yield extract_tool_return(self.response_model, json_output, True)
        if json_output:
            json_output = json_output[: json_output.rfind("}") + 1]
        self.constructed_response_model = extract_tool_return(
            self.response_model, json_output, False
        )
        yield self.constructed_response_model

    def __aiter__(self) -> AsyncGenerator[_ResponseModelT, None]:
        """Iterates over the stream and extracts structured outputs."""

        async def generator():
            json_output = ""
            async for chunk, _ in self.stream:
                json_output += chunk.content
                if json_output and json_output[0] != "{":
                    try:
                        json_start = json_output.index("{")
                        json_output = json_output[json_start:]
                    except ValueError:
                        json_output = ""
                if chunk.model is not None:
                    self.stream.model = chunk.model
                if json_output:
                    yield extract_tool_return(self.response_model, json_output, True)
            if json_output:
                json_output = json_output[: json_output.rfind("}") + 1]
            self.constructed_response_model = extract_tool_return(
                self.response_model, json_output, False
            )
            yield self.constructed_response_model

        return generator()


_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_P = ParamSpec("_P")


def structured_stream_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TStream: type[BaseStream],
    TToolType: type[_BaseToolT],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseChunkT],
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Iterable[_ResponseModelT],
    ]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Awaitable[AsyncIterable[_ResponseModelT]],
    ]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        Iterable[_ResponseModelT] | Awaitable[AsyncIterable[_ResponseModelT]],
    ]:
        def handle_stream(
            stream: Generator[_ResponseChunkT, None, None],
            tool_types: list[type[_BaseToolT]] | None,
        ):
            for chunk in stream:
                call_response_chunk = TCallResponseChunk(chunk=chunk)
                json_output = get_json_output(call_response_chunk, json_mode)
                call_response_chunk_type = type(call_response_chunk)
                setattr(call_response_chunk_type, "content", json_output)
                yield call_response_chunk_type(chunk=chunk), None

        async def handle_stream_async(
            stream: AsyncGenerator[_ResponseChunkT, None],
            tool_types: list[type[_BaseToolT]] | None,
        ):
            async for chunk in stream:
                call_response_chunk = TCallResponseChunk(chunk=chunk)
                json_output = get_json_output(call_response_chunk, json_mode)
                call_response_chunk_type = type(call_response_chunk)
                setattr(call_response_chunk_type, "content", json_output)
                yield call_response_chunk_type(chunk=chunk), None

        stream_decorator = stream_factory(
            TCallResponse=TCallResponse,
            TStream=TStream,
            setup_call=setup_call,
            handle_stream=handle_stream,
            handle_stream_async=handle_stream_async,
        )

        tool = setup_extract_tool(response_model, TToolType)
        is_async = inspect.iscoroutinefunction(fn)
        stream_decorator_kwargs = {
            "model": model,
            "tools": [tool],
            "json_mode": json_mode,
            "client": client,
            "call_params": call_params,
        }

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> Iterable[_ResponseModelT]:
            assert SetupCall.fn_is_sync(fn)
            return BaseStructuredStream[_ResponseModelT](
                stream=stream_decorator(fn=fn, **stream_decorator_kwargs)(
                    *args, **kwargs
                ),
                response_model=response_model,
            )

        @wraps(fn)
        async def inner_async(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> AsyncIterable[_ResponseModelT]:
            assert SetupCall.fn_is_async(fn)
            return BaseStructuredStream[_ResponseModelT](
                stream=await stream_decorator(fn=fn, **stream_decorator_kwargs)(
                    *args, **kwargs
                ),
                response_model=response_model,
            )

        return inner_async if is_async else inner

    return decorator



"""A module for interacting with Weaviate vectorstores."""

from .types import WeaviateSettings, WeaviateParams
from .vectorstores import WeaviateVectorStore

__all__ = ["WeaviateSettings", "WeaviateParams", "WeaviateVectorStore"]



"""Handles streaming content and tools from the Gemini API."""

from collections.abc import AsyncGenerator, Generator

from google.generativeai.types import GenerateContentResponse  # type: ignore

from ..call_response_chunk import GeminiCallResponseChunk
from ..tool import GeminiTool


def handle_stream(
    stream: Generator[GenerateContentResponse, None, None],
    tool_types: list[type[GeminiTool]] | None = None,
) -> Generator[tuple[GeminiCallResponseChunk, None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed.

    Note: gemini does not currently support streaming tools.
    """
    for chunk in stream:
        yield GeminiCallResponseChunk(chunk=chunk), None


async def handle_stream_async(
    stream: AsyncGenerator[GenerateContentResponse, None],
    tool_types: list[type[GeminiTool]] | None = None,
) -> AsyncGenerator[tuple[GeminiCallResponseChunk, None], None]:
    """
    Async iterator over the stream and constructs tools as they are streamed.

    Note: gemini does not currently support streaming tools.
    """
    async for chunk in stream:
        yield GeminiCallResponseChunk(chunk=chunk), None



from openai.types import Embedding
from openai.types.create_embedding_response import CreateEmbeddingResponse

from ..base.embedding_response import BaseEmbeddingResponse


class OpenAIEmbeddingResponse(BaseEmbeddingResponse[CreateEmbeddingResponse]):
    """A convenience wrapper around the OpenAI `CreateEmbeddingResponse` response."""

    @property
    def embeddings(self) -> list[list[float]]:
        """Returns the raw embeddings."""
        embeddings_model: list[Embedding] = [
            embedding for embedding in self.response.data
        ]
        return [embedding.embedding for embedding in embeddings_model]



"""The Mirascope Mistral Module."""

from ._call import mistral_call
from ._call import mistral_call as call
from .call_params import MistralCallParams
from .call_response import MistralCallResponse
from .call_response_chunk import MistralCallResponseChunk
from .dynamic_config import MistralDynamicConfig
from .stream import MistralStream
from .tool import MistralTool

__all__ = [
    "call",
    "MistralDynamicConfig",
    "MistralCallParams",
    "MistralCallResponse",
    "MistralCallResponseChunk",
    "MistralStream",
    "MistralTool",
    "mistral_call",
]



"""Integrations with third party libraries."""

from contextlib import suppress

from .middleware_factory import middleware_decorator

with suppress(ImportError):
    from . import logfire as logfire

with suppress(ImportError):
    from . import langfuse as langfuse

with suppress(ImportError):
    from . import otel as otel

with suppress(ImportError):
    from . import tenacity as tenacity


__all__ = ["langfuse", "logfire", "middleware_decorator", "otel", "tenacity"]



"""The `extract_factory` method for generating provider specific create decorators."""

import inspect
from functools import wraps
from typing import Awaitable, Callable, ParamSpec, TypeVar, overload

from pydantic import BaseModel, ValidationError

from ._create import create_factory
from ._utils import (
    BaseType,
    GetJsonOutput,
    SetupCall,
    extract_tool_return,
    setup_extract_tool,
)
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_ResponseChunkT = TypeVar("_ResponseChunkT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
_P = ParamSpec("_P")


def extract_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TToolType: type[BaseTool],
    setup_call: SetupCall[
        _BaseClientT,
        _BaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _BaseToolT,
    ],
    get_json_output: GetJsonOutput[_BaseCallResponseT],
):
    """Returns the wrapped function with the provider specific interfaces."""
    create_decorator = create_factory(
        TCallResponse=TCallResponse, setup_call=setup_call
    )

    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, _ResponseModelT | _ParsedOutputT]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P, Awaitable[_ResponseModelT | _ParsedOutputT]
    ]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
        model: str,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[
        _P,
        _ResponseModelT | _ParsedOutputT | Awaitable[_ResponseModelT | _ParsedOutputT],
    ]:
        tool = setup_extract_tool(response_model, TToolType)
        is_async = inspect.iscoroutinefunction(fn)
        create_decorator_kwargs = {
            "model": model,
            "tools": [tool],
            "output_parser": None,
            "json_mode": json_mode,
            "client": client,
            "call_params": call_params,
        }

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert SetupCall.fn_is_sync(fn)
            call_response = create_decorator(fn=fn, **create_decorator_kwargs)(
                *args, **kwargs
            )
            json_output = get_json_output(call_response, json_mode)
            try:
                output = extract_tool_return(response_model, json_output, False)
            except ValidationError as e:
                e._response = call_response  # type: ignore
                raise e
            if isinstance(output, BaseModel):
                output._response = call_response  # type: ignore
            return output if not output_parser else output_parser(output)  # type: ignore

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _ResponseModelT:
            assert SetupCall.fn_is_async(fn)
            call_response = await create_decorator(fn=fn, **create_decorator_kwargs)(
                *args, **kwargs
            )
            json_output = get_json_output(call_response, json_mode)
            try:
                output = extract_tool_return(response_model, json_output, False)
            except ValidationError as e:
                e._response = call_response  # type: ignore
                raise e
            if isinstance(output, BaseModel):
                output._response = call_response  # type: ignore
            return output if not output_parser else output_parser(output)  # type: ignore

        return inner_async if is_async else inner

    return decorator



"""This module contains the function to extract the return value of a tool."""

from typing import TypeVar

import jiter
from pydantic import BaseModel

from .._partial import partial
from ._base_type import BaseType, is_base_type
from ._convert_base_type_to_base_tool import convert_base_type_to_base_tool

_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)


def extract_tool_return(
    response_model: type[_ResponseModelT],
    json_output: str | object,
    allow_partial: bool,
) -> _ResponseModelT:
    json_obj = (
        jiter.from_json(
            json_output.encode(),
            partial_mode="trailing-strings" if allow_partial else "off",
        )
        if isinstance(json_output, str)
        else json_output
    )
    if is_base_type(response_model):
        temp_model = convert_base_type_to_base_tool(response_model, BaseModel)  # type: ignore
        if allow_partial:
            return partial(temp_model).model_validate(json_obj).value  # type: ignore
        return temp_model.model_validate(json_obj).value  # type: ignore

    if allow_partial:
        return partial(response_model).model_validate(json_obj)  # type: ignore
    return response_model.model_validate(json_obj)  # type: ignore



"""This module contains the `CohereCallResponseChunk` class."""

from cohere import StreamedChatResponse_StreamEnd, StreamedChatResponse_StreamStart
from cohere.types import (
    ApiMetaBilledUnits,
    ChatStreamEndEventFinishReason,
    StreamedChatResponse,
    StreamedChatResponse_TextGeneration,
)
from pydantic import SkipValidation

from ..base import BaseCallResponseChunk


class CohereCallResponseChunk(
    BaseCallResponseChunk[
        SkipValidation[StreamedChatResponse], ChatStreamEndEventFinishReason
    ]
):
    '''A convenience wrapper around the Cohere `ChatCompletionChunk` streamed chunks.

    When calling the Cohere API using a function decorated with `cohere_call` and
    `stream` set to `True`, the stream will contain `CohereResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.cohere import cohere_call

    @cohere_call(model="gpt-4o", stream=True)
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    stream = recommend_book("fantasy")  # response is an `CohereStream`
    for chunk in stream:
        print(chunk.content, end="", flush=True)
    #> Sure! I would recommend...
    ```
    '''

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        if isinstance(self.chunk, StreamedChatResponse_TextGeneration):
            return self.chunk.text
        return ""

    @property
    def finish_reasons(self) -> list[ChatStreamEndEventFinishReason] | None:
        """Returns the finish reasons of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            return [self.chunk.finish_reason]
        return None

    @property
    def model(self) -> str | None:
        """Returns the name of the response model.

        Cohere does not return model, so we return None
        """
        return None

    @property
    def id(self) -> str | None:
        """Returns the id of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamStart):
            return self.chunk.generation_id
        elif isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            return self.chunk.response.generation_id
        return None

    @property
    def usage(self) -> ApiMetaBilledUnits | None:
        """Returns the usage of the response."""
        if isinstance(self.chunk, StreamedChatResponse_StreamEnd):
            if self.chunk.response.meta:
                return self.chunk.response.meta.billed_units
        return None

    @property
    def input_tokens(self) -> float | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.input_tokens
        return None

    @property
    def output_tokens(self) -> float | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.output_tokens
        return None



"""Utility for converting a model model into a base tool."""

import inspect
from abc import update_abstractmethods
from typing import Any, TypeVar, cast

from pydantic import BaseModel, create_model

from ._default_tool_docstring import DEFAULT_TOOL_DOCSTRING

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def convert_base_model_to_base_tool(
    model: type[BaseModel], base: type[BaseToolT]
) -> type[BaseToolT]:
    """Converts a `BaseModel` schema to a `BaseToolT` type.

    By adding a docstring (if needed) and passing on fields and field information in
    dictionary format, a Pydantic `BaseModel` can be converted into an `BaseTool` for
    performing extraction.

    Args:
        schema: The `BaseModel` schema to convert.

    Returns:
        The constructed `BaseModelT` type.
    """
    field_definitions = {
        field_name: (field_info.annotation, field_info)
        for field_name, field_info in model.model_fields.items()
    }
    tool_type = create_model(
        f"{model.__name__}",
        __base__=base,
        __doc__=model.__doc__ if model.__doc__ else DEFAULT_TOOL_DOCSTRING,
        **cast(dict[str, Any], field_definitions),
    )
    for name, value in inspect.getmembers(model):
        if not hasattr(tool_type, name) or name in ["_name", "_description", "call"]:
            setattr(tool_type, name, value)
    return update_abstractmethods(tool_type)



"""This module contains the `AnthropicCallResponse` class."""

from anthropic.types import Message, MessageParam, ToolResultBlockParam, Usage
from pydantic import computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import AnthropicCallParams
from .dynamic_config import AnthropicDynamicConfig
from .tool import AnthropicTool


class AnthropicCallResponse(
    BaseCallResponse[
        Message,
        AnthropicTool,
        AnthropicDynamicConfig,
        MessageParam,
        AnthropicCallParams,
        MessageParam,
    ]
):
    '''A convenience wrapper around the Anthropic `Message` response.

    When calling the Anthropic API using a function decorated with `anthropic_call`, the
    response will be an `AnthropicCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.anthropic import anthropic_call

    @anthropic_call(model="claude-3-5-sonnet-20240620")
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    response = recommend_book("fantasy")  # response is `AnthropicCallResponse` instance
    print(response.content)
    #> Sure! I would recommend...
    ```
    '''

    _provider = "anthropic"

    @property
    def content(self) -> str:
        """Returns the string text of the 0th text block."""
        block = self.response.content[0]
        return block.text if block.type == "text" else ""

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(self.response.stop_reason)]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> Usage:
        """Returns the usage of the message."""
        return self.response.usage

    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.input_tokens

    @property
    def output_tokens(self) -> int:
        """Returns the number of output tokens."""
        return self.usage.output_tokens

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> MessageParam:
        """Returns the assistants's response as a message parameter."""
        return self.response.model_dump(include={"content", "role"})  # type: ignore

    @computed_field
    @property
    def tools(self) -> list[AnthropicTool] | None:
        """Returns any available tool calls as their `AnthropicTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        if not self.tool_types:
            return None

        extracted_tools = []
        for content in self.response.content:
            if content.type != "tool_use":
                continue
            for tool_type in self.tool_types:
                if content.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(content))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> AnthropicTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[AnthropicTool, str]]
    ) -> list[MessageParam]:
        """Returns the tool message parameters for tool call results."""
        return [
            {
                "role": "user",
                "content": [
                    ToolResultBlockParam(
                        tool_use_id=tool.tool_call.id,
                        type="tool_result",
                        content=[{"text": output, "type": "text"}],
                    )
                    for tool, output in tools_and_outputs
                ],
            }
        ]



"""This module defines the function return type for functions as LLM calls."""

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import CohereCallParams

CohereDynamicConfig = BaseDynamicConfig[BaseMessageParam, CohereCallParams]
'''The function return type for functions wrapped with the `cohere_call` decorator.

Attributes:
    computed_fields: The computed fields to use in the prompt template.
    messages: The messages to send to the Cohere API. If provided, `computed_fields`
        will be ignored.
    call_params: The call parameters to use when calling the Cohere API. These will
        override any call parameters provided to the decorator.

Example:

```python
from mirascope.core.cohere import CohereDynamicConfig, cohere_call

@cohere_call(model="gpt-4o")
def recommend_book(genre: str) -> CohereDynamicConfig:
    """Recommend a {capitalized_genre} book."""
    reeturn {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
'''



from typing import Any

from pydantic import BaseModel


class Document(BaseModel):
    """A document to be added to the vectorstore."""

    id: str
    text: str
    metadata: dict[str, Any] | None = None



"""Utility for converting `BaseMessageParam` to `ContentsType`"""

import io

import PIL.Image
from google.generativeai.types import ContentDict

from ...base import BaseMessageParam


def convert_message_params(
    message_params: list[BaseMessageParam | ContentDict],
) -> list[ContentDict]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif (role := message_param.role) == "system":
            content = message_param.content
            if not isinstance(message_param.content, str):
                raise ValueError(
                    "System message content must be a single text string."
                )  # pragma: no cover
            converted_message_params += [
                {
                    "role": "user",
                    "parts": [message_param.content],
                },
                {
                    "role": "model",
                    "parts": ["Ok! I will adhere to this system message."],
                },
            ]
        elif isinstance((content := message_param.content), str):
            converted_message_params.append({"role": "user", "parts": [content]})
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.text)
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/webp",
                        "image/heic",
                        "image/heif",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. "
                            "Gemini currently only supports JPEG, PNG, WebP, HEIC, "
                            "and HEIF images."
                        )
                    image = PIL.Image.open(io.BytesIO(part.image))
                    converted_content.append(image)
                else:
                    if part.media_type not in [
                        "audio/wav",
                        "audio/mp3",
                        "audio/aiff",
                        "audio/aac",
                        "audio/ogg",
                        "audio/flac",
                    ]:
                        raise ValueError(
                            f"Unsupported audio media type: {part.media_type}. "
                            "Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, "
                            "and FLAC audio file types."
                        )
                    converted_content.append(
                        {"mime_type": part.media_type, "data": part.audio}
                    )
            converted_message_params.append({"role": role, "parts": converted_content})
    return converted_message_params



"""This module contains utility functions for base types."""

import inspect
from enum import Enum
from typing import Annotated, Any, Literal, TypeGuard, Union, get_origin

BaseType = str | int | float | bool | bytes | list | set | tuple | dict


def is_base_type(type_: Any) -> TypeGuard[type[BaseType]]:
    """Check if a type is a base type."""
    base_types = {str, int, float, bool, bytes, list, set, tuple, dict}
    return (
        (inspect.isclass(type_) and issubclass(type_, Enum))
        or type_ in base_types
        or get_origin(type_) in base_types.union({Literal, Union, Annotated})
    )



"""Utilities for handling a stream of messages."""

from collections.abc import AsyncGenerator, Generator

import jiter
from anthropic.types import MessageStreamEvent, ToolUseBlock

from ..call_response_chunk import AnthropicCallResponseChunk
from ..tool import AnthropicTool


def _handle_chunk(
    buffer: str,
    chunk: MessageStreamEvent,
    current_tool_call: ToolUseBlock,
    current_tool_type: type[AnthropicTool] | None,
    tool_types: list[type[AnthropicTool]] | None,
) -> tuple[
    str,
    AnthropicTool | None,
    ToolUseBlock,
    type[AnthropicTool] | None,
]:
    """Handles a chunk of the stream."""
    if not tool_types:
        return buffer, None, current_tool_call, current_tool_type

    if chunk.type == "content_block_stop" and current_tool_type and buffer:
        current_tool_call.input = jiter.from_json(buffer.encode())
        return (
            "",
            current_tool_type.from_tool_call(current_tool_call),
            ToolUseBlock(id="", input={}, name="", type="tool_use"),
            None,
        )

    if chunk.type == "content_block_start" and isinstance(
        chunk.content_block, ToolUseBlock
    ):
        content_block = chunk.content_block
        current_tool_type = None
        for tool_type in tool_types:
            if tool_type._name() == content_block.name:
                current_tool_type = tool_type
                break
        if current_tool_type is None:
            raise RuntimeError(
                f"Unknown tool type in stream: {content_block.name}."
            )  # pragma: no cover
        return (
            "",
            None,
            ToolUseBlock(
                id=content_block.id, input={}, name=content_block.name, type="tool_use"
            ),
            current_tool_type,
        )

    if chunk.type == "content_block_delta" and chunk.delta.type == "input_json_delta":
        buffer += chunk.delta.partial_json

    return buffer, None, current_tool_call, current_tool_type


def handle_stream(
    stream: Generator[MessageStreamEvent, None, None],
    tool_types: list[type[AnthropicTool]] | None,
) -> Generator[tuple[AnthropicCallResponseChunk, AnthropicTool | None], None, None]:
    """Iterator over the stream and constructs tools as they are streamed."""
    current_tool_call = ToolUseBlock(id="", input={}, name="", type="tool_use")
    current_tool_type, buffer = None, ""
    for chunk in stream:
        buffer, tool, current_tool_call, current_tool_type = _handle_chunk(
            buffer, chunk, current_tool_call, current_tool_type, tool_types
        )
        yield AnthropicCallResponseChunk(chunk=chunk), tool


async def handle_stream_async(
    stream: AsyncGenerator[MessageStreamEvent, None],
    tool_types: list[type[AnthropicTool]] | None,
) -> AsyncGenerator[tuple[AnthropicCallResponseChunk, AnthropicTool | None], None]:
    current_tool_call = ToolUseBlock(id="", input={}, name="", type="tool_use")
    current_tool_type, buffer = None, ""
    async for chunk in stream:
        buffer, tool, current_tool_call, current_tool_type = _handle_chunk(
            buffer, chunk, current_tool_call, current_tool_type, tool_types
        )
        yield AnthropicCallResponseChunk(chunk=chunk), tool



"""Mirascope x Logfire Integration utils"""

from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Generator,
)

import logfire
from pydantic import BaseModel

from mirascope.core.base._utils._base_type import BaseType

from ...core.base import BaseCallResponse, _utils
from ...core.base._stream import BaseStream
from ...core.base._structured_stream import BaseStructuredStream
from ...core.base.metadata import Metadata


@contextmanager
def custom_context_manager(
    fn: Callable,
) -> Generator[logfire.LogfireSpan, Any, None]:
    metadata: Metadata = _utils.get_metadata(fn, None)
    tags = metadata.get("tags", [])
    with logfire.with_settings(custom_scope_suffix="mirascope", tags=list(tags)).span(
        fn.__name__
    ) as logfire_span:
        yield logfire_span


def get_call_response_span_data(result: BaseCallResponse) -> dict:
    output: dict[str, Any] = {}
    if cost := result.cost:
        output["cost"] = cost
    if input_tokens := result.input_tokens:
        output["input_tokens"] = input_tokens
    if output_tokens := result.output_tokens:
        output["output_tokens"] = output_tokens
    if content := result.content:
        output["content"] = content
    return {
        "async": False,
        "call_params": result.call_params,
        "call_kwargs": result.call_kwargs,
        "model": result.model,
        "provider": result._provider,
        "prompt_template": result.prompt_template,
        "template_variables": result.fn_args,
        "messages": result.messages,
        "response_data": result.response,
        "output": output,
    }


def get_tool_calls(result: BaseCallResponse) -> list[dict] | None:
    if tools := result.tools:
        tool_calls = [
            {
                "function": {
                    "arguments": tool.args,
                    "name": tool._name(),
                }
            }
            for tool in tools
        ]
        return tool_calls
    return None


def handle_call_response(
    result: BaseCallResponse, fn: Callable, logfire_span: logfire.LogfireSpan | None
):
    if logfire_span is None:
        return
    span_data = get_call_response_span_data(result)
    span_data["async"] = False
    tool_calls = get_tool_calls(result)
    if tool_calls:
        span_data["output"]["tool_calls"] = tool_calls
    logfire_span.set_attributes(span_data)


def handle_stream(
    stream: BaseStream,
    fn: Callable,
    logfire_span: logfire.LogfireSpan | None,
):
    handle_call_response(stream.construct_call_response(), fn, logfire_span)


def set_response_model_output(
    result: BaseModel | BaseType, output: dict[str, Any] = {}
) -> None:
    if isinstance(result, BaseModel):
        output["response_model"] = {
            "name": result.__class__.__name__,
            "arguments": result.model_dump(),
        }
    else:
        output["content"] = result


def handle_response_model(
    result: BaseModel | BaseType, fn: Callable, logfire_span: logfire.LogfireSpan | None
):
    if logfire_span is None:
        return
    span_data: dict[str, Any] = {"output": {}, "async": False}
    if isinstance(result, BaseModel):  # type: ignore
        response: BaseCallResponse = result._response  # type: ignore
        span_data |= get_call_response_span_data(response)
    set_response_model_output(result, span_data["output"])
    logfire_span.set_attributes(span_data)


def get_structured_stream_span_data(result: BaseStructuredStream) -> dict:
    span_data = get_call_response_span_data(result.stream.construct_call_response())
    output = span_data.get("output", {})
    response_model = result.constructed_response_model
    if isinstance(response_model, BaseModel):
        output["response_model"] = {
            "name": response_model.__class__.__name__,
            "arguments": response_model.model_dump(),
        }
    else:
        output["content"] = response_model
    span_data["output"] = output

    return span_data


def handle_structured_stream(
    result: BaseStructuredStream, fn: Callable, logfire_span: logfire.LogfireSpan | None
):
    if logfire_span is None:
        return
    span_data = get_structured_stream_span_data(result)
    span_data["async"] = False
    logfire_span.set_attributes(span_data)


async def handle_call_response_async(
    result: BaseCallResponse, fn: Callable, logfire_span: logfire.LogfireSpan | None
):
    if logfire_span is None:
        return
    span_data = get_call_response_span_data(result)
    span_data["async"] = True
    tool_calls = get_tool_calls(result)
    if tool_calls:
        span_data["output"]["tool_calls"] = tool_calls
    logfire_span.set_attributes(span_data)


async def handle_stream_async(
    stream: BaseStream,
    fn: Callable,
    logfire_span: logfire.LogfireSpan | None,
):
    await handle_call_response_async(stream.construct_call_response(), fn, logfire_span)


async def handle_response_model_async(
    result: BaseModel | BaseType,
    fn: Callable,
    logfire_span: logfire.LogfireSpan | None,
):
    if logfire_span is None:
        return
    span_data: dict[str, Any] = {"output": {}, "async": True}
    if isinstance(result, BaseModel):  # type: ignore
        response: BaseCallResponse = result._response  # type: ignore
        span_data |= get_call_response_span_data(response)
    set_response_model_output(result, span_data["output"])
    logfire_span.set_attributes(span_data)


async def handle_structured_stream_async(
    result: BaseStructuredStream,
    fn: Callable,
    logfire_span: logfire.LogfireSpan | None,
):
    if logfire_span is None:
        return
    span_data = get_structured_stream_span_data(result)
    span_data["async"] = True
    logfire_span.set_attributes(span_data)



"""Get the JSON output from a completion response."""

import json

from cohere.types import StreamedChatResponse_ToolCallsGeneration

from ..call_response import CohereCallResponse
from ..call_response_chunk import CohereCallResponseChunk


def get_json_output(
    response: CohereCallResponse | CohereCallResponseChunk, json_mode: bool
) -> str:
    """Get the JSON output from a completion response."""
    if isinstance(response, CohereCallResponse):
        if json_mode and response.content:
            return response.content
        elif response.response.tool_calls:
            return json.dumps(response.response.tool_calls[0].parameters)
        raise ValueError("No tool call or JSON object found in response.")
    else:
        # raise ValueError("Cohere does not support structured streaming... :(")
        if json_mode:
            return response.content
        elif (
            isinstance(response.chunk, StreamedChatResponse_ToolCallsGeneration)
            and (tool_calls := response.chunk.tool_calls)
            and (parameters := tool_calls[0].parameters)
        ):
            return json.dumps(parameters)
        return ""



"""Cohere utilities for decorator factories."""

from ._calculate_cost import calculate_cost
from ._convert_message_params import convert_message_params
from ._get_json_output import get_json_output
from ._handle_stream import handle_stream, handle_stream_async
from ._setup_call import setup_call

__all__ = [
    "calculate_cost",
    "convert_message_params",
    "get_json_output",
    "handle_stream",
    "handle_stream_async",
    "setup_call",
]



"""The base type in a function as an LLM call to return for dynamic configuration."""

from typing import Any, Callable, Generic, TypeVar

from typing_extensions import NotRequired, TypedDict

from .call_params import BaseCallParams
from .metadata import Metadata
from .tool import BaseTool

_MessageParamT = TypeVar("_MessageParamT", bound=Any)
_CallParamsT = TypeVar("_CallParamsT", bound=BaseCallParams)


class DynamicConfigBase(TypedDict):
    metadata: NotRequired[Metadata]
    computed_fields: NotRequired[dict[str, Any | list[Any] | list[list[Any]]]]
    tools: NotRequired[list[type[BaseTool] | Callable]]


class DynamicConfigMessages(DynamicConfigBase, Generic[_MessageParamT]):
    messages: NotRequired[list[_MessageParamT]]


class DynamicConfigCallParams(DynamicConfigBase, Generic[_CallParamsT]):
    call_params: NotRequired[_CallParamsT]


class DynamicConfigFull(DynamicConfigBase, Generic[_MessageParamT, _CallParamsT]):
    messages: NotRequired[list[_MessageParamT]]
    call_params: NotRequired[_CallParamsT]


BaseDynamicConfig = (
    DynamicConfigBase
    | DynamicConfigMessages[_MessageParamT]
    | DynamicConfigCallParams[_CallParamsT]
    | DynamicConfigFull[_MessageParamT, _CallParamsT]
    | None
)
"""The base type in a function as an LLM call to return for dynamic configuration."""



"""This module contains the `setup_extract_tool` function."""

from typing import TypeVar

from pydantic import BaseModel

from ._base_type import BaseType, is_base_type
from ._convert_base_model_to_base_tool import convert_base_model_to_base_tool
from ._convert_base_type_to_base_tool import convert_base_type_to_base_tool

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def setup_extract_tool(
    response_model: type[BaseModel] | type[BaseType], tool_type: type[BaseToolT]
) -> type[BaseToolT]:
    if is_base_type(response_model):
        return convert_base_type_to_base_tool(response_model, tool_type)  # type: ignore
    return convert_base_model_to_base_tool(response_model, tool_type)  # type: ignore



from typing import Literal

from httpx import Timeout
from openai._types import Body, Headers, Query

from ..base.embedding_params import BaseEmbeddingParams


class OpenAIEmbeddingParams(BaseEmbeddingParams):
    model: str = "text-embedding-3-small"
    encoding_format: Literal["float", "base64"] | None = None
    user: str | None = None
    # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
    # The extra values given here take precedence over values defined on the client or passed to this method.
    extra_headers: Headers | None = None
    extra_query: Query | None = None
    extra_body: Body | None = None
    timeout: float | Timeout | None = None
